"""Implementation of DataUnits and Data class to store calibration routines outputs."""

import re
from abc import abstractmethod

import numpy as np
import pandas as pd
import pint_pandas

from qibocal.config import raise_error


class AbstractData:
    """Base class for the implementation of `DataUnits` and `Data`."""

    def __init__(self, name=None):
        if name is None:
            self.name = "data"
        else:
            self.name = name

        self.df = pd.DataFrame()
        self.quantities = None

    def __add__(self, data):
        self.df = pd.concat([self.df, data.df], ignore_index=True)
        return self

    @abstractmethod
    def add(self, data):  # pragma: no cover
        """Add row to `AbstractData` dataframe."""
        raise_error(NotImplementedError)

    def __len__(self):
        """Computes the length of the data."""
        return len(self.df)

    @classmethod
    def load_data(
        cls, folder, subfolder, routine, data_format, name
    ):  # pragma: no cover
        """Load data from specific format."""
        raise_error(NotImplementedError)

    @abstractmethod
    def to_csv(self, path):  # pragma: no cover
        """Save data in csv file.

        Args:
            path (str): Path containing output folder."""
        raise_error(NotImplementedError)

    def to_pickle(self, path):
        """Save data in pickel file.

        Args:
            path (str): Path containing output folder."""
        self.df.to_pickle(f"{path}/{self.name}.pkl")


class DataUnits(AbstractData):
    """Class to store the data measured during the calibration routines.
    It is a wrapper to a pandas DataFrame with units of measure from the Pint
    library.

    Args:
        quantities (dict): dictionary containing additional quantities that the user
                        may save other than the pulse sequence output. The keys are the name of the
                        quantities and the corresponding values are the units of measure.
        options (list): list containing additional values to be saved.
    """

    def __init__(self, name=None, quantities=None, options=None):
        super().__init__(name=name)

        self._df = pd.DataFrame(
            {
                "MSR": pd.Series(dtype="pint[V]"),
                "i": pd.Series(dtype="pint[V]"),
                "q": pd.Series(dtype="pint[V]"),
                "phase": pd.Series(dtype="pint[rad]"),
            }
        )
        self.quantities = {"MSR": "V", "i": "V", "q": "V", "phase": "rad"}
        self.options = []

        if quantities is not None:
            self.quantities.update(quantities)
            for quantity, unit in quantities.items():
                self.df.insert(0, quantity, pd.Series(dtype=f"pint[{unit}]"))

        if options is not None:
            self.options = options
            for option in options:
                self.df.insert(  # pylint: disable=E1101
                    0, option, pd.Series(dtype=object)
                )

        from pint import UnitRegistry

        self.ureg = UnitRegistry()

    @property
    def df(self):
        """Dataframe attribute."""
        return self._df

    @df.setter
    def df(self, df):
        """Set df attribute.

        Args:
            df (pd.DataFrame): pandas DataFrame. Every key should have the following form:
                               ``<name>[<unit>]``.
        """
        if isinstance(df, pd.DataFrame):
            self._df = df
        else:
            raise_error(TypeError, f"{type(df)} is not a pd.DataFrame.")

    def load_data_from_dict(self, data: dict):
        """Load dataframe from dictionary.

        Args:
            data (dict): dictionary containing the data to be loaded.
                        Every key should have the following form:
                        ``<name>[<unit>]``.
        """
        processed_data = {}
        for key, values in data.items():
            if "[" in key:
                name = key.split("[")[0]
                unit = re.search(r"\[([A-Za-z0-9_]+)\]", key).group(1)
                processed_data[name] = pd.Series(
                    data=(np.array(values) * self.ureg(unit)), dtype=f"pint[{unit}]"
                )
            else:
                processed_data[key] = pd.Series(data=(values), dtype=object)
        self._df = pd.DataFrame(processed_data)

    def add_data_from_dict(self, data: dict):
        """Add the contents of a dictionary to the data of the dataframe.
        Args:
            df (dict): dictionary containing the data to be added.
        """
        processed_data = {}
        for key, values in data.items():
            if "[" in key:
                name = key.split("[")[0]
                unit = re.search(r"\[([A-Za-z0-9_]+)\]", key).group(1)
                processed_data[name] = pd.Series(
                    data=(np.array(values) * self.ureg(unit)), dtype=f"pint[{unit}]"
                )
            else:
                processed_data[key] = pd.Series(data=(values), dtype=object)
        self._df = pd.concat(
            [self._df, pd.DataFrame(processed_data)], ignore_index=True
        )

    def add(self, data):
        """Add a row to `DataUnits`.

        Args:
            data (dict): dictionary containing the data to be added.
                        Every key should have the following form:
                        ``<name>[<unit>]``.
        """
        l = len(self)

        for key, value in data.items():
            if "[" in key:
                name = key.split("[")[0]
                unit = re.search(r"\[([A-Za-z0-9_]+)\]", key).group(1)
                # TODO: find a better way to do this
                self.df.loc[l, name] = np.array(value) * self.ureg(unit)
            else:
                self.df.loc[l, key] = value

    def get_values(self, key, unit=None):
        """Get values of a quantity in specified units.

        Args:
            quantity (str): Quantity to get the values of.
            unit (str): Unit of the returned values.

        Returns:
            ``pd.Series`` with the quantity values in the given units.
        """
        if unit is None:
            return self.df[key]

        return self.df[key].pint.to(unit).pint.magnitude

    @classmethod
    def load_data(cls, folder, subfolder, routine, data_format, name):
        """Load data from specific format.

        Args:
            folder (path): path to the output folder from which the data will be loaded
            subfolder (path): subfolder name
            routine (str): calibration routine data to be loaded
            data_format (str): data format. Possible choices are 'csv' and 'pickle'
            name (str): file's name without extension
        Returns:
            data (``DataUnits``): dataset object with the loaded data.
        Example:
            see the method ``load_data`` in class ``Data``
        """
        obj = cls()
        if data_format == "csv":
            file = f"{folder}/{subfolder}/{routine}/{name}.csv"
            obj.df = pd.read_csv(file, header=[0, 1])
            obj.df.pop("Unnamed: 0_level_0")
            quantities_label = []
            obj.options = []
            for column in obj.df.columns:  # pylint: disable=E1101
                if "Unnamed" not in column[1]:
                    quantities_label.append(column[0])
                else:
                    obj.options.append(column[0])
            quantities_df = obj.df[quantities_label].pint.quantify()
            options_df = obj.df[obj.options]
            options_df.columns = options_df.columns.droplevel(1)
            obj.df = pd.concat([quantities_df, options_df], axis=1)
        elif data_format == "pickle":
            file = f"{folder}/data/{routine}/{name}.pkl"
            obj.df = pd.read_pickle(file)
        else:
            raise_error(ValueError, f"Cannot load data using {data_format} format.")

        return obj

    def to_csv(self, path):
        """Save data in csv file.

        Args:
            path (str): Path containing output folder.
        Example:
            .. testcode::

                import os
                from qibocal.data import DataUnits
                import numpy as np
                data = DataUnits()
                length = 3
                folder = "foo"
                # create directory
                if not os.path.isdir(folder):
                    os.mkdir(folder)
                # generate random dataset
                for l in range(length):
                    msr, i, q, phase = np.random.rand(4)
                    pulse_sequence_result = {
                        "MSR[V]": msr,
                        "i[V]": i,
                        "q[V]": q,
                        "phase[rad]]": phase,
                    }
                    data.add({**pulse_sequence_result})
                data.to_csv(folder)

            .. testcleanup::

                import shutil
                if os.path.isdir("foo"):
                    shutil.rmtree("foo")
        """
        data = self.df[list(self.quantities)].pint.dequantify()
        firsts = data.index.get_level_values(None)
        data[self.options] = self.df[self.options].loc[firsts].values
        data.to_csv(f"{path}/{self.name}.csv")


class Data(AbstractData):
    """Class to store the data obtained from calibration routines.
    It is a wrapper to a pandas DataFrame.

    Args:
        quantities (list): list of quantities to be saved.
    """

    def __init__(self, name=None, quantities=None):
        super().__init__(name=name)

        if quantities is not None:
            self.quantities = quantities
            for quantity in quantities:
                self.df.insert(0, quantity, pd.Series(dtype=object))

    @property
    def df(self):
        """Dataframe attribute."""
        return self._df

    @df.setter
    def df(self, data):
        """Set df attribute.

        Args:
            df (pd.DataFrame):
        """
        if isinstance(data, pd.DataFrame):
            self._df = data

    def load_data_from_dict(self, data: dict):
        """Set df attribute.

        Args:
            df (dict): dictionary containing the data to be added.
        Example:
            .. testcode::

                from qibocal.data import Data
                data = Data()
                test = {
                    "int": [1, 2, 3],
                    "float": [3.0, 4.0, 5.0],
                    "string": ["one", "two", "three"],
                    "bool": [True, False, True],
                }
                data.load_data_from_dict(test)
                print(data.df)
            .. testoutput::

                  int float string   bool
                0   1   3.0    one   True
                1   2   4.0    two  False
                2   3   5.0  three   True
        """
        processed_data = {}
        for key, values in data.items():
            processed_data[key] = pd.Series(data=(values), dtype=object)
        self._df = pd.DataFrame(processed_data)

    def add(self, data):
        """Add a row to data.

        Args:
            data (dict): dictionary containing the data to be added.
                        Every key should have the following form:
                        ``<name>[<unit>]``.
        """
        l = len(self)
        for key, value in data.items():
            self.df.loc[l, key] = value

    def get_values(self, quantity):
        """Get values of a quantity.

        Args:
            quantity (str): Quantity to get the values of.

        Returns:
            ``pd.Series`` with the quantity values in the given units.
        """
        return self.df[quantity].values

    @classmethod
    def load_data(cls, folder, subfolder, routine, data_format, name):
        """Load data from specific format.

        Args:
            folder (path): path to the output folder from which the data will be loaded
            subfolder (path): subfolder name
            routine (str): calibration routine data to be loaded
            data_format (str): data format. Possible choices are 'csv' and 'pickle'
            name (str): file's name without extension

        Returns:
            data (``Data``): data object with the loaded data.

        Example:
            .. testcode::

                from qibocal.data import Data
                import os
                folder = "test_folder/test_subfolder/test_routine"
                length = 3
                if not os.path.isdir(folder):
                    os.makedirs(folder)
                # create a dataset
                data = Data()
                for i in range(length):
                    data.add(
                        {
                            "int": int(i),
                            "float": float(i),
                            "string": str(f"hello{i}"),
                            "bool": bool(i),
                        }
                    )
                # save the dataset in csv format
                data.to_csv(folder)
                # upload the dataset
                data_upload = Data().load_data("test_folder", "test_subfolder", "test_routine", "csv", "data")

            .. testcleanup::

                import shutil
                shutil.rmtree("test_folder")

        """

        obj = cls()
        if data_format == "csv":
            file = f"{folder}/{subfolder}/{routine}/{name}.csv"
            obj.df = pd.read_csv(file)
            obj.df.pop("Unnamed: 0")
        elif data_format == "pickle":
            file = f"{folder}/{subfolder}/{routine}/{name}.pkl"
            obj.df = pd.read_pickle(file)
        else:
            raise_error(ValueError, f"Cannot load data using {data_format} format.")

        return obj

    def to_csv(self, path):
        """Save data in csv file.

        Args:
            path (str): Path containing output folder.
        Example:
            .. testcode::

                import shutil
                from qibocal.data import Data
                import numpy as np
                import os
                data = Data()
                length = 3
                folder = "foo"
                # create directory
                if not os.path.isdir(folder):
                    os.mkdir(folder)
                # generate a dataset
                data = Data()
                for i in range(length):
                    data.add(
                        {
                            "int": int(i),
                            "float": float(i),
                            "string": str(f"hello{i}"),
                            "bool": bool(i),
                        }
                    )
                data.to_csv(folder)

            .. testcleanup::

                import shutil
                if os.path.isdir("foo"):
                    shutil.rmtree("foo")
        """
        self.df.to_csv(f"{path}/{self.name}.csv")



import os
import pathlib
import subprocess

import yaml
from flask import Flask, redirect, render_template, request

from qibocal import __version__
from qibocal.cli.builders import ReportBuilder

server = Flask(__name__)


@server.route("/")
@server.route("/data/<path>")
def page(path=None):
    folders = [
        folder
        for folder in reversed(sorted(os.listdir(os.getcwd())))
        if os.path.isdir(folder) and "meta.yml" in os.listdir(folder)
    ]

    report = []
    if path is not None:
        try:
            report.append(ReportBuilder(path))
        except (FileNotFoundError, TypeError):
            pass

    return render_template(
        "template.html",
        version=__version__,
        folders=folders,
        reports=report,
    )


@server.route("/", methods=["GET", "POST"])
@server.route("/data/<path>", methods=["GET", "POST"])
def qq_compare():
    if request.method == "POST":
        if request.form["qq-compare"] == "compare":
            list_of_folders = ""
            selected_folders = request.form.getlist("list_of_folders")
            for folder in selected_folders:
                list_of_folders = list_of_folders + folder + " "

            # execute shell command qq-compare
            command = "qq-compare " + list_of_folders
            try:
                command_output = subprocess.check_output([command], shell=True)

            except subprocess.CalledProcessError as e:
                # return "An error occurred while trying to generate qq-report.<br>Please check folder compatility."
                # return redirect("/", code=302)
                folders = [
                    folder
                    for folder in reversed(sorted(os.listdir(os.getcwd())))
                    if os.path.isdir(folder) and "meta.yml" in os.listdir(folder)
                ]
                report = []
                return render_template(
                    "template.html",
                    version=__version__,
                    folders=folders,
                    reports=report,
                    error="An error occurred while trying to generate qq-report. Please check folder compatility.",
                )

            return redirect("/data/qq-compare", code=302)

        elif request.form["qq-compare"] == "combine":
            folders = [
                folder
                for folder in reversed(sorted(os.listdir(os.getcwd())))
                if os.path.isdir(folder) and "meta.yml" in os.listdir(folder)
            ]

            list_of_reports = []
            selected_folders = request.form.getlist("list_of_folders")
            for folder in selected_folders:
                try:
                    list_of_reports.append(ReportBuilder(folder))
                except (FileNotFoundError, TypeError):
                    pass

            return render_template(
                "template.html",
                version=__version__,
                folders=folders,
                reports=list_of_reports,
            )



import numpy as np
import pandas as pd
import pytest
from plotly.graph_objects import Figure
from qibo import gates, models

from qibocal.calibrations.niGSC import simulfilteredrb
from qibocal.calibrations.niGSC.basics import noisemodels, utils


@pytest.fixture
def depths():
    return [0, 1, 5, 10, 30]


@pytest.fixture
def nshots():
    return 13


@pytest.mark.parametrize("nqubits", [1, 2, 3])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("qubits", [[0], [0, 1]])
def test_experiment(nqubits: int, depths: list, runs: int, nshots: int, qubits: list):
    if max(qubits) > nqubits - 1:
        pass
    else:
        myfactory1 = simulfilteredrb.ModuleFactory(nqubits, depths * runs)
        myexperiment1 = simulfilteredrb.ModuleExperiment(myfactory1, nshots=nshots)
        myexperiment1.perform(myexperiment1.execute)
        assert isinstance(myexperiment1.data, list)
        assert isinstance(myexperiment1.data[0], dict)
        for count, datarow in enumerate(myexperiment1.data):
            assert len(datarow.keys()) == 2
            assert isinstance(datarow["samples"], np.ndarray)
            assert len(datarow["samples"]) == nshots
            assert isinstance(datarow["depth"], int)
            assert datarow["depth"] == depths[count % len(depths)]
            if not datarow["depth"]:
                assert np.array_equal(
                    datarow["samples"], np.zeros(datarow["samples"].shape)
                )
        assert isinstance(myexperiment1.dataframe, pd.DataFrame)


@pytest.mark.parametrize("nqubits", [1, 3])
@pytest.mark.parametrize("noise_params", [[0.1, 0.2, 0.1], [0.03, 0.17, 0.05]])
def test_experiment_withnoise(nqubits: int, noise_params):
    nshots = 512
    depths = [0, 5, 10]
    runs = 2
    # Build the noise model.
    noise = noisemodels.PauliErrorOnUnitary(*noise_params)
    # Test exectue an experiment.
    myfactory1 = simulfilteredrb.ModuleFactory(nqubits, depths * runs)
    circuit_list = list(myfactory1)
    myfaultyexperiment = simulfilteredrb.ModuleExperiment(
        circuit_list, nshots=nshots, noise_model=noise
    )
    myfaultyexperiment.perform(myfaultyexperiment.execute)
    experiment1 = simulfilteredrb.ModuleExperiment(circuit_list, nshots=nshots)
    experiment1.perform(experiment1.execute)
    experiment12 = simulfilteredrb.ModuleExperiment(circuit_list, nshots=nshots)
    experiment12.perform(experiment12.execute)
    for datarow_faulty, datarow2, datarow3 in zip(
        myfaultyexperiment.data, experiment1.data, experiment12.data
    ):
        if not datarow_faulty["depth"]:
            assert np.array_equal(datarow_faulty["samples"], datarow2["samples"])
            assert np.array_equal(datarow2["samples"], datarow3["samples"])
        else:
            probs1 = utils.probabilities(datarow_faulty["samples"])
            probs2 = utils.probabilities(datarow2["samples"])
            probs3 = utils.probabilities(datarow3["samples"])
            # Since the error channel maps the state into the maximally mixed state,
            # if the state is already maximally mixed, they will not differ.
            # So check if a probabiliy entry is zero.
            if np.any(probs2 == 0):
                assert np.linalg.norm(probs1 - probs2) > np.linalg.norm(probs2 - probs3)


def test_filterfunction():
    """Test if the filter function works, without noise."""

    nqubits = 2
    nshots = 3000
    d = 2
    # Steal the class method for calculating clifford unitaries.
    from qibo.quantum_info.random_ensembles import random_clifford

    # The first parameter is self, set it to None since it is not needed.
    g1_matrix = random_clifford(1)
    g1 = gates.Unitary(g1_matrix, 0)
    g2_matrix = random_clifford(1)
    g2 = gates.Unitary(g2_matrix, 0)
    g3_matrix = random_clifford(1)
    g3 = gates.Unitary(g3_matrix, 1)
    g4_matrix = random_clifford(1)
    g4 = gates.Unitary(g4_matrix, 1)
    # Calculate the ideal unitary and the ideal outcomes.
    g21 = g2_matrix @ g1_matrix
    g43 = g4_matrix @ g3_matrix
    ideal1 = g21 @ np.array([[1], [0]])
    ideal2 = g43 @ np.array([[1], [0]])
    # Build the circuit with the ideal unitaries.
    c = models.Circuit(nqubits)
    c.add([g1, g3, g2, g4])
    c.add(gates.M(0, 1))
    # Execute the circuit and get the samples.
    samples = c(nshots=nshots).samples()
    # Initiate the variables to store the four irrep signals.
    a0, a1, a2, a3 = 0, 0, 0, 0
    for s in samples:
        # lambda = (0,0)
        a0 += 1 / d**2
        # lambda = (0,1)
        a1 += 1 / d * np.abs(ideal2[s[1]]) ** 2 - 1 / d**2
        # lambda = (1,0)
        a2 += 1 / d * np.abs(ideal1[s[0]]) ** 2 - 1 / d**2
        # lambda = (1,1)
        a3 += (
            np.abs(ideal1[s[0]]) ** 2 * np.abs(ideal2[s[1]]) ** 2
            - 1 / d * np.abs(ideal1[s[0]]) ** 2
            - 1 / d * np.abs(ideal2[s[1]]) ** 2
            + 1 / d**2
        )
    a0 *= 1 / (nshots)
    a1 *= (d + 1) / (nshots)
    a2 *= (d + 1) / (nshots)
    a3 *= (d + 1) ** 2 / (nshots)
    # Now do the same but with an experiment, use a list with only
    # the prebuild circuit (build it again because it was already executed).
    # No noise.
    c = models.Circuit(nqubits)
    c.add([g1, g3, g2, g4])
    c.add(gates.M(0, 1))
    experiment = simulfilteredrb.ModuleExperiment([c], nshots=nshots)
    experiment.perform(experiment.execute)
    # Compute and get the filtered signals.
    experiment.perform(simulfilteredrb.filter_function)
    # Compare the above calculated filtered signals and the signals
    # computed with the crosstalkrb method.
    assert isinstance(experiment.data, list)
    assert np.isclose(a0, experiment.data[0]["irrep0"])
    assert np.isclose(a1, experiment.data[0]["irrep1"])
    assert np.isclose(a2, experiment.data[0]["irrep2"])
    assert np.isclose(a3, experiment.data[0]["irrep3"])


@pytest.mark.parametrize("nqubits", [2, 3])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("noise_params", [[0.1, 0.1, 0.1], [0.4, 0.2, 0.01]])
def test_post_processing(
    nqubits: int, depths: list, runs: int, nshots: int, noise_params: list
):
    # Build the noise model.
    noise = noisemodels.PauliErrorOnUnitary(*noise_params)
    # Test exectue an experiment.
    myfactory1 = simulfilteredrb.ModuleFactory(nqubits, list(depths) * runs)
    myfaultyexperiment = simulfilteredrb.ModuleExperiment(
        myfactory1, nshots=nshots, noise_model=noise
    )
    myfaultyexperiment.perform(myfaultyexperiment.execute)
    simulfilteredrb.post_processing_sequential(myfaultyexperiment)
    aggr_df = simulfilteredrb.get_aggregational_data(myfaultyexperiment)
    assert (
        len(aggr_df) == 2**nqubits
        and aggr_df.index[0] == "irrep0"
        and aggr_df.index[1] == "irrep1"
    )
    assert "depth" in aggr_df.columns
    assert "data" in aggr_df.columns
    assert "2sigma" in aggr_df.columns
    assert "fit_func" in aggr_df.columns
    assert "popt" in aggr_df.columns
    assert "perr" in aggr_df.columns


def test_build_report():
    depths = [1, 5, 10, 15, 20, 25]
    nshots = 128
    runs = 10
    nqubits = 1
    noise_params = [0.01, 0.1, 0.05]
    # Build the noise model.
    noise = noisemodels.PauliErrorOnUnitary(*noise_params)
    # Test exectue an experiment.
    myfactory1 = simulfilteredrb.ModuleFactory(nqubits, depths * runs)
    myfaultyexperiment = simulfilteredrb.ModuleExperiment(
        myfactory1, nshots=nshots, noise_model=noise
    )
    myfaultyexperiment.perform(myfaultyexperiment.execute)
    simulfilteredrb.post_processing_sequential(myfaultyexperiment)
    aggr_df = simulfilteredrb.get_aggregational_data(myfaultyexperiment)
    report_figure = simulfilteredrb.build_report(myfaultyexperiment, aggr_df)
    assert isinstance(report_figure, Figure)



from itertools import product

import numpy as np
import pandas as pd
import pytest
from plotly.graph_objects import Figure
from qibo import gates
from qibo.noise import NoiseModel

from qibocal.calibrations.niGSC import standardrb
from qibocal.calibrations.niGSC.basics import noisemodels, utils


def theoretical_outcome(noise_model: NoiseModel) -> float:
    """Take the used noise model acting on unitaries and calculates the
    effective depolarizing parameter.

    Args:
        experiment (Experiment): Experiment which executed the simulation.
        noisemodel (NoiseModel): Applied noise model.

    Returns:
        (float): The effective depolarizing parameter of given error.
    """

    # TODO This has to be more systematic. Delete it from the branch which will be merged.
    # Check for correctness of noise model and gate independence.
    errorkeys = noise_model.errors.keys()
    assert len(errorkeys) == 1 and list(errorkeys)[0] == gates.Unitary
    # Extract the noise acting on unitaries and turn it into the associated
    # error channel.
    error = noise_model.errors[gates.Unitary][0]
    errorchannel = error.channel(0, *error.options)
    # Calculate the effective depolarizing parameter.
    return utils.effective_depol(errorchannel)


@pytest.fixture
def depths():
    return [0, 1, 5, 10, 30]


@pytest.fixture
def nshots():
    return 127


@pytest.mark.parametrize("nqubits", [1, 2])
@pytest.mark.parametrize("runs", [1, 3])
def test_experiment(nqubits: int, depths: list, runs: int, nshots: int):
    # Test execute an experiment.
    myfactory1 = standardrb.ModuleFactory(nqubits, list(depths) * runs)
    myexperiment1 = standardrb.ModuleExperiment(myfactory1, nshots=nshots)
    assert myexperiment1.name == "StandardRB"
    myexperiment1.perform(myexperiment1.execute)
    assert isinstance(myexperiment1.data, list)
    assert isinstance(myexperiment1.data[0], dict)
    for count, datarow in enumerate(myexperiment1.data):
        assert len(datarow.keys()) == 2
        assert isinstance(datarow["samples"], np.ndarray)
        assert len(datarow["samples"]) == nshots
        assert isinstance(datarow["depth"], int)
        assert datarow["depth"] == depths[count % len(depths)]
        assert np.array_equal(datarow["samples"], np.zeros(datarow["samples"].shape))
    assert isinstance(myexperiment1.dataframe, pd.DataFrame)


@pytest.mark.parametrize("nqubits", [1, 3])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("noise_params", [[0.1, 0.1, 0.1], [0.02, 0.3, 0.07]])
def test_experiment_withnoise(
    nqubits: int, depths: list, runs: int, nshots: int, noise_params: list
):
    # Build the noise model.
    noise = noisemodels.PauliErrorOnUnitary(*noise_params)
    # Test exectue an experiment.
    myfactory1 = standardrb.ModuleFactory(nqubits, list(depths) * runs)
    myfaultyexperiment = standardrb.ModuleExperiment(
        myfactory1, nshots=nshots, noise_model=noise
    )
    myfaultyexperiment.perform(myfaultyexperiment.execute)
    assert isinstance(myfaultyexperiment.data, list)
    assert isinstance(myfaultyexperiment.data[0], dict)
    for count, datarow in enumerate(myfaultyexperiment.data):
        assert len(datarow.keys()) == 2
        assert isinstance(datarow["samples"], np.ndarray)
        assert len(datarow["samples"]) == nshots
        assert isinstance(datarow["depth"], int)
        assert datarow["depth"] == depths[count % len(depths)]
        # If there are no executed gates (other than the measurement) or only one
        # the probability that no error occured is too high.
        if datarow["depth"] > 2:
            assert not np.array_equal(
                datarow["samples"], np.zeros(datarow["samples"].shape)
            )
    assert isinstance(myfaultyexperiment.dataframe, pd.DataFrame)


@pytest.mark.parametrize("nqubits", [2, 3])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("qubits", [[0], [0, 1]])
def test_embed_circuit(nqubits: int, depths: list, runs: int, qubits: list):
    nshots = 2
    myfactory1 = standardrb.ModuleFactory(nqubits, list(depths) * runs, qubits=qubits)
    test_list = list(product(qubits))
    test_list.append(tuple(qubits))
    for circuit in myfactory1:
        assert circuit.nqubits == nqubits
        for gate in circuit.queue:
            assert gate._target_qubits in test_list
    myexperiment1 = standardrb.ModuleExperiment(myfactory1, nshots=nshots)
    myexperiment1.perform(myexperiment1.execute)


@pytest.mark.parametrize("nqubits", [2, 3])
@pytest.mark.parametrize("runs", [1, 3])
def test_utils_probs_and_noisy_execution(
    nqubits: int, depths: list, runs: int, nshots: int
):
    noise_params = [0.0001, 0.001, 0.0005]
    # Build the noise model.
    noise = noisemodels.PauliErrorOnUnitary(*noise_params)
    # Test exectue an experiment.
    myfactory1 = standardrb.ModuleFactory(nqubits, list(depths) * runs)
    myfaultyexperiment = standardrb.ModuleExperiment(
        myfactory1, nshots=nshots, noise_model=noise
    )
    myfaultyexperiment.perform(myfaultyexperiment.execute)
    myfaultyexperiment.perform(standardrb.groundstate_probabilities)
    probs = utils.probabilities(myfaultyexperiment.extract("samples"))
    assert probs.shape == (runs * len(depths), 2**nqubits)
    assert np.allclose(np.sum(probs, axis=1), 1)
    for probsarray in probs:
        if probsarray[0] < 1.0:
            assert np.all(
                np.greater_equal(probsarray[0] * np.ones(len(probsarray)), probsarray)
            )
        else:
            assert probsarray[0] == 1.0


@pytest.mark.parametrize("nqubits", [2, 3])
@pytest.mark.parametrize("runs", [1, 3])
def test_post_processing(nqubits: int, depths: list, runs: int, nshots: int):
    noise_params = [0.01, 0.3, 0.14]
    # Build the noise model.
    noise = noisemodels.PauliErrorOnUnitary(*noise_params)
    # Test exectue an experiment.
    myfactory1 = standardrb.ModuleFactory(nqubits, list(depths) * runs)
    myfaultyexperiment = standardrb.ModuleExperiment(
        myfactory1, nshots=nshots, noise_model=noise
    )
    myfaultyexperiment.perform(myfaultyexperiment.execute)
    standardrb.post_processing_sequential(myfaultyexperiment)
    probs = utils.probabilities(myfaultyexperiment.extract("samples"))
    ground_probs = probs[:, 0]
    test_ground_probs = myfaultyexperiment.extract("groundstate probability")
    assert np.allclose(ground_probs, test_ground_probs)
    aggr_df = standardrb.get_aggregational_data(myfaultyexperiment)
    assert len(aggr_df) == 1 and aggr_df.index[0] == "groundstate probability"
    assert "depth" in aggr_df.columns
    assert "data" in aggr_df.columns
    assert "2sigma" in aggr_df.columns
    assert "fit_func" in aggr_df.columns
    assert "popt" in aggr_df.columns
    assert "perr" in aggr_df.columns


def test_build_report():
    depths = [1, 5, 10, 15, 20, 25]
    nshots = 1024
    runs = 5
    nqubits = 1
    noise_params = [0.01, 0.1, 0.05]
    # Build the noise model.
    noise = noisemodels.PauliErrorOnUnitary(*noise_params)
    # Test exectue an experiment.
    myfactory1 = standardrb.ModuleFactory(nqubits, depths * runs)
    myfaultyexperiment = standardrb.ModuleExperiment(
        myfactory1, nshots=nshots, noise_model=noise
    )
    myfaultyexperiment.perform(myfaultyexperiment.execute)
    standardrb.post_processing_sequential(myfaultyexperiment)
    aggr_df = standardrb.get_aggregational_data(myfaultyexperiment)
    assert (
        theoretical_outcome(noise) - aggr_df.popt[0]["p"]
        < 2 * aggr_df.perr[0]["p_err"] + theoretical_outcome(noise) * 0.05
    )
    report_figure = standardrb.build_report(myfaultyexperiment, aggr_df)
    assert isinstance(report_figure, Figure)



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import flipping_fit


@plot("MSR vs Flips", plots.flips_msr)
def flipping(
    platform: AbstractPlatform,
    qubits: dict,
    nflips_max,
    nflips_step,
    software_averages=1,
    points=10,
):
    r"""
    The flipping experiment correct the delta amplitude in the qubit drive pulse. We measure a qubit after applying
    a Rx(pi/2) and N flips (Rx(pi) rotations). After fitting we can obtain the delta amplitude to refine pi pulses.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        nflips_max (int): Maximum number of flips introduced in each sequence
        nflips_step (int): Scan range step for the number of flippings
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **flips[dimensionless]**: Number of flips applied in the current execution
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **amplitude_correction_factor**: Pi pulse correction factor
            - **corrected_amplitude**: Corrected pi pulse amplitude
            - **popt0**: p0
            - **popt1**: p1
            - **popt2**: p2
            - **popt3**: p3
            - **qubit**: The qubit being tested
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a DataUnits object to store MSR, phase, i, q and the number of flips
    data = DataUnits(
        name="data",
        quantities={"flips": "dimensionless"},
        options=["qubit", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        # sweep the parameter
        for flips in range(0, nflips_max, nflips_step):
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data
                # calculate and save fit
                yield flipping_fit(
                    data,
                    x="flips[dimensionless]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    pi_pulse_amplitudes={
                        q: qubits[qubit].pi_pulse_amplitude for q in qubits
                    },
                    labels=["amplitude_correction_factor", "corrected_amplitude"],
                )

            # create a sequence of pulses for the experiment
            sequence = PulseSequence()
            ro_pulses = {}
            for qubit in qubits:
                RX90_pulse = platform.create_RX90_pulse(qubit, start=0)
                sequence.add(RX90_pulse)
                # execute sequence RX(pi/2) - [RX(pi) - RX(pi)] from 0...flips times - RO
                start1 = RX90_pulse.duration
                for j in range(flips):
                    RX_pulse1 = platform.create_RX_pulse(qubit, start=start1)
                    start2 = start1 + RX_pulse1.duration
                    RX_pulse2 = platform.create_RX_pulse(qubit, start=start2)
                    sequence.add(RX_pulse1)
                    sequence.add(RX_pulse2)
                    start1 = start2 + RX_pulse2.duration

                # add ro pulse at the end of the sequence
                ro_pulses[qubit] = platform.create_qubit_readout_pulse(
                    qubit, start=start1
                )
                sequence.add(ro_pulses[qubit])

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            for ro_pulse in ro_pulses.values():
                # average msr, phase, i and q over the number of shots defined in the runcard
                r = results[ro_pulse.serial].to_dict()
                r.update(
                    {
                        "flips[dimensionless]": flips,
                        "qubit": ro_pulse.qubit,
                        "iteration": iteration,
                    }
                )
                data.add(r)
            count += 1
    yield data

    yield flipping_fit(
        data,
        x="flips[dimensionless]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        pi_pulse_amplitudes={q: qubits[qubit].pi_pulse_amplitude for q in qubits},
        labels=["amplitude_correction_factor", "corrected_amplitude"],
    )



import re

import numpy as np
from scipy.special import mathieu_a, mathieu_b


def lorenzian(frequency, amplitude, center, sigma, offset):
    # http://openafox.com/science/peak-function-derivations.html
    return (amplitude / np.pi) * (
        sigma / ((frequency - center) ** 2 + sigma**2)
    ) + offset


def rabi(x, p0, p1, p2, p3, p4):
    # A fit to Superconducting Qubit Rabi Oscillation
    #   Offset                       : p[0]
    #   Oscillation amplitude        : p[1]
    #   Period    T                  : 1/p[2]
    #   Phase                        : p[3]
    #   Arbitrary parameter T_2      : 1/p[4]
    return p0 + p1 * np.sin(2 * np.pi * x * p2 + p3) * np.exp(-x * p4)


def ramsey(x, p0, p1, p2, p3, p4):
    # A fit to Superconducting Qubit Rabi Oscillation
    #   Offset                       : p[0]
    #   Oscillation amplitude        : p[1]
    #   DeltaFreq                    : p[2]
    #   Phase                        : p[3]
    #   Arbitrary parameter T_2      : 1/p[4]
    return p0 + p1 * np.sin(x * p2 + p3) * np.exp(-x * p4)


def exp(x, *p):
    return p[0] - p[1] * np.exp(-1 * x * p[2])


def flipping(x, p0, p1, p2, p3):
    # A fit to Flipping Qubit oscillation
    # Epsilon?? shoule be Amplitude : p[0]
    # Offset                        : p[1]
    # Period of oscillation         : p[2]
    # phase for the first point corresponding to pi/2 rotation   : p[3]
    return np.sin(x * 2 * np.pi / p2 + p3) * p0 + p1


def cos(x, p0, p1, p2, p3):
    # Offset                  : p[0]
    # Amplitude               : p[1]
    # Period                  : p[2]
    # Phase                   : p[3]
    return p0 + p1 * np.cos(2 * np.pi * x / p2 + p3)


def line(x, p0, p1):
    # Slope                   : p[0]
    # Intercept               : p[1]
    return p0 * x + p1


def parse(key):
    name = key.split("[")[0]
    unit = re.search(r"\[([A-Za-z0-9_]+)\]", key).group(1)
    return name, unit


def G_f_d(x, p0, p1, p2):
    # Current offset:          : p[0]
    # 1/I_0, Phi0=Xi*I_0       : p[1]
    # Junction asymmetry d     : p[2]
    G = np.sqrt(
        np.cos(np.pi * (x - p0) * p1) ** 2
        + p2**2 * np.sin(np.pi * (x - p0) * p1) ** 2
    )
    return np.sqrt(G)


def freq_r_transmon(x, p0, p1, p2, p3, p4, p5):
    # Current offset:                                      : p[0]
    # 1/I_0, Phi0=Xi*I_0                                   : p[1]
    # Junction asymmetry d                                 : p[2]
    # f_q0/f_rh, f_q0 = Qubit frequency at zero flux       : p[3]
    # Qubit-resonator coupling at zero magnetic flux, g_0  : p[4]
    # High power resonator frequency, f_rh                 : p[5]
    return p5 + p4**2 * G_f_d(x, p0, p1, p2) / (p5 - p3 * p5 * G_f_d(x, p0, p1, p2))


def kordering(m, ng=0.4999):
    # Ordering function sorting the eigenvalues |m,ng> for the Schrodinger equation for the
    # Cooper pair box circuit in the phase basis.
    a1 = (round(2 * ng + 1 / 2) % 2) * (round(ng) + 1 * (-1) ** m * divmod(m + 1, 2)[0])
    a2 = (round(2 * ng - 1 / 2) % 2) * (round(ng) - 1 * (-1) ** m * divmod(m + 1, 2)[0])
    return a1 + a2


def mathieu(index, x):
    # Mathieu's characteristic value a_index(x).
    if index < 0:
        dummy = mathieu_b(-index, x)
    else:
        dummy = mathieu_a(index, x)
    return dummy


def freq_q_mathieu(x, p0, p1, p2, p3, p4, p5=0.499):
    # Current offset:                                      : p[0]
    # 1/I_0, Phi0=Xi*I_0                                   : p[1]
    # Junction asymmetry d                                 : p[2]
    # Charging energy E_C                                  : p[3]
    # Josephson energy E_J                                 : p[4]
    # Effective offset charge ng                           : p[5]
    index1 = int(2 * (p5 + kordering(1, p5)))
    index0 = int(2 * (p5 + kordering(0, p5)))
    p4 = p4 * G_f_d(x, p0, p1, p2)
    return p3 * (mathieu(index1, -p4 / (2 * p3)) - mathieu(index0, -p4 / (2 * p3)))


def freq_r_mathieu(x, p0, p1, p2, p3, p4, p5, p6, p7=0.499):
    # High power resonator frequency, f_rh                 : p[0]
    # Qubit-resonator coupling at zero magnetic flux, g_0  : p[1]
    # Current offset:                                      : p[2]
    # 1/I_0, Phi0=Xi*I_0                                   : p[3]
    # Junction asymmetry d                                 : p[4]
    # Charging energy E_C                                  : p[5]
    # Josephson energy E_J                                 : p[6]
    # Effective offset charge ng                           : p[7]
    G = G_f_d(x, p2, p3, p4)
    f_q = freq_q_mathieu(x, p2, p3, p4, p5, p6, p7)
    f_r = p0 + p1**2 * np.sqrt(G) / (p0 - f_q)
    return f_r



import warnings

import numpy as np
import pytest

from qibocal.calibrations.niGSC.basics import fitting


def test_1expfitting():
    successes = 0
    number_runs = 50
    for _ in range(number_runs):
        x = np.sort(np.random.choice(np.linspace(0, 30, 50), size=20, replace=False))
        A, B = np.random.uniform(0.1, 0.99, size=2)
        f = np.random.uniform(0.4, 0.99)
        y = A * f**x + B
        assert np.allclose(fitting.exp1B_func(x, A, f, B), y)
        # Distort ``y`` a bit.
        y_dist = y + np.random.randn(len(y)) * 0.005
        popt, perr = fitting.fit_exp1B_func(x, y_dist)
        successes += np.all(
            np.abs(np.array(popt) - [A, f, B])
            < 2 * np.array(perr) + 0.05 * np.array([A, f, B]),
        )
    assert successes >= number_runs * 0.8

    successes = 0
    number_runs = 50
    for _ in range(number_runs):
        x = np.sort(np.random.choice(np.linspace(0, 30, 50), size=20, replace=False))
        A = np.random.uniform(0.1, 0.99)
        f = np.random.uniform(0.4, 0.99)
        y = A * f**x
        # Distort ``y`` a bit.
        y_dist = y + np.random.randn(len(y)) * 0.005
        popt, perr = fitting.fit_exp1_func(x, y_dist)
        successes += np.all(
            np.abs(np.array(popt) - [A, f])
            < 2 * np.array(perr) + 0.05 * np.array([A, f])
        )
    assert successes >= number_runs * 0.8

    x = np.sort(np.random.choice(np.linspace(-5, 5, 50), size=20, replace=False))
    y = np.zeros(len(x)) + 0.75
    assert np.array_equal(
        np.array(fitting.fit_exp1B_func(x, y)), np.array(((0.75, 1.0, 0), (0, 0, 0)))
    )
    assert np.array_equal(
        np.array(fitting.fit_exp1_func(x, y)), np.array(((0.75, 1.0), (0, 0)))
    )
    # At least once the algorithm shall not find a fit:
    didnt_getit = 0
    didnt_getitB = 0
    for _ in range(20):
        x = np.sort(np.random.choice(np.linspace(0, 15, 50), size=50, replace=False))
        y_dist = np.e ** (-((x - 5) ** 2) * 10) + np.random.randn(len(x)) * 0.1
        popt1, perr1 = fitting.fit_exp1_func(x, y_dist, p0=[-100])
        didnt_getit += not (np.all(np.array([*popt1, *perr1]), 0))
        popt, perr = fitting.fit_exp1B_func(x, y_dist, p0=[-100, 0.01])
        didnt_getitB += not (np.all(np.array([*popt, *perr]), 0))
    assert didnt_getit >= 1 and didnt_getitB >= 1


def test_exp2_fitting():
    successes = 0
    number_runs = 50
    for count in range(number_runs):
        x = np.arange(0, 50)
        A1, A2 = np.random.uniform(0.1, 0.99, size=2)
        if not count % 3:
            f1, f2 = np.random.uniform(0.1, 0.5, size=2) * 1j + np.random.uniform(
                0.1, 0.99, size=2
            )
        else:
            f1, f2 = np.random.uniform(0.1, 0.99, size=2)
        y = A1 * f1**x + A2 * f2**x
        assert np.allclose(fitting.exp2_func(x, A1, A2, f1, f2), y)
        # Distort ``y`` a bit.
        y_dist = y + np.random.uniform(-1, 1, size=len(y)) * 0.001
        popt, perr = fitting.fit_exp2_func(x, y_dist)
        worked = np.all(
            np.logical_or(
                np.allclose(np.array(popt), [A2, A1, f2, f1], atol=0.05, rtol=0.1),
                np.allclose(np.array(popt), [A1, A2, f1, f2], atol=0.05, rtol=0.1),
            )
        )
        if not worked:
            np.allclose(
                popt[0] * popt[2] ** x + popt[1] * popt[3] ** x,
                y_dist,
                atol=0.01,
                rtol=0.01,
            )
            worked = True
        successes += worked
    # This is a pretty bad rate. The ESPRIT algorithm has to be optimized.
    assert successes >= number_runs * 0.4

    with pytest.raises(ValueError):
        x = np.array([1, 2, 3, 5])
        A1, A2, f1, f2 = np.random.uniform(0.1, 0.99, size=4)
        y = A1 * f1**x + A2 * f2**x
        # Distort ``y`` a bit.
        y_dist = y + np.random.uniform(-1, 1, size=len(y)) * 0.001
        popt, perr = fitting.fit_exp2_func(x, y_dist)



"""Routine-specific method for post-processing data acquired."""
from functools import partial

import lmfit
import numpy as np
from scipy.optimize import curve_fit

from qibocal.config import log
from qibocal.data import Data
from qibocal.fitting.utils import (
    cos,
    exp,
    flipping,
    freq_q_mathieu,
    freq_r_mathieu,
    freq_r_transmon,
    line,
    lorenzian,
    parse,
    rabi,
    ramsey,
)


def lorentzian_fit(data, x, y, qubits, resonator_type, labels, fit_file_name=None):
    r"""
    Fitting routine for resonator/qubit spectroscopy.
    The used model is

    .. math::

        y = \frac{A}{\pi} \Big[ \frac{\sigma}{(f-f_0)^2 + \sigma^2} \Big] + y_0.

    Args:

    Args:
        data (`DataUnits`): dataset for the fit
        x (str): name of the input values for the Lorentzian model
        y (str): name of the output values for the Lorentzian model
        qubits (list): A list with the IDs of the qubits
        resonator_type (str): the type of readout resonator ['3D', '2D']
        labels (list of str): list containing the lables of the quantities computed by this fitting method.

            -   When using ``resonator_spectroscopy`` the expected labels are [`readout_frequency`, `peak voltage`], where `readout_frequency` is the estimated frequency of the resonator, and `peak_voltage` the peak of the Lorentzian

            -   when using ``qubit_spectroscopy`` the expected labels are [`drive_frequency`, `peak voltage`], where `drive_frequency` is the estimated frequency of the qubit

        fit_file_name (str): file name, ``None`` is the default value.

    Returns:

        A ``Data`` object with the following keys

            - **labels[0]**: peak voltage
            - **labels[1]**: frequency
            - **popt0**: Lorentzian's amplitude
            - **popt1**: Lorentzian's center
            - **popt2**: Lorentzian's sigma
            - **popt3**: Lorentzian's offset
            - **qubit**: The qubit being tested

    Example:

        In the code below, a noisy Lorentzian dataset is implemented and then the ``lorentzian_fit`` method is applied.

            .. testcode::

                import numpy as np
                from qibocal.data import DataUnits
                from qibocal.fitting.methods import lorentzian_fit
                from qibocal.fitting.utils import lorenzian
                import matplotlib.pyplot as plt

                name = "test"
                nqubits = 1
                label = "drive_frequency"
                amplitude = -1
                center = 2
                sigma = 3
                offset = 4

                # generate noisy Lorentzian

                x = np.linspace(center - 10, center + 10, 100)
                noisy_lorentzian = (
                    lorenzian(x, amplitude, center, sigma, offset)
                    + amplitude * np.random.randn(100) * 0.5e-2
                )

                # Initialize data and evaluate the fit

                data = DataUnits(quantities={"frequency": "Hz"}, options=["qubit", "iteration"])

                mydict = {"frequency[Hz]": x, "MSR[V]": noisy_lorentzian, "qubit": 0, "iteration" : 0}

                data.load_data_from_dict(mydict)

                fit = lorentzian_fit(
                    data,
                    "frequency[Hz]",
                    "MSR[V]",
                    qubits = [0],
                    resonator_type='3D',
                    labels=[label, "peak_voltage", "intermediate_freq"],
                    fit_file_name=name,
                )

                fit_params = [fit.get_values(f"popt{i}") for i in range(4)]
                fit_data = lorenzian(x,*fit_params)

                # Plot

                #fig = plt.figure(figsize = (10,5))
                #plt.scatter(x,noisy_lorentzian,label="data",s=10,color = 'darkblue',alpha = 0.9)
                #plt.plot(x,fit_data, label = "fit", color = 'violet', linewidth = 3, alpha = 0.4)
                #plt.xlabel('frequency (Hz)')
                #plt.ylabel('MSR (Volt)')
                #plt.legend()
                #plt.title("Data fit")
                #plt.grid()
                #plt.show()

            The following plot shows the resulting output:

            .. image:: lorentzian_fit_result.png
                :align: center

    """
    if fit_file_name == None:
        data_fit = Data(
            name=f"fits",
            quantities=[
                "popt0",
                "popt1",
                "popt2",
                "popt3",
                labels[0],
                labels[1],
                "qubit",
            ],
        )
    else:
        data_fit = Data(
            name=fit_file_name,
            quantities=[
                "popt0",
                "popt1",
                "popt2",
                "popt3",
                labels[0],
                labels[1],
                "qubit",
            ],
        )
    for qubit in qubits:
        qubit_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby("frequency", as_index=False)
            .mean()
        )
        frequencies_keys = parse(x)
        voltages_keys = parse(y)
        frequencies = (
            qubit_data[frequencies_keys[0]].pint.to(frequencies_keys[1]).pint.magnitude
        )
        voltages = qubit_data[voltages_keys[0]].pint.to(voltages_keys[1]).pint.magnitude

        # Create a lmfit model for fitting equation defined in resonator_peak
        model_Q = lmfit.Model(lorenzian)

        # Guess parameters for Lorentzian max or min
        if (resonator_type == "3D" and "readout_frequency" in labels[0]) or (
            resonator_type != "3D" and labels[0] == "drive_frequency"
        ):
            guess_center = frequencies[
                np.argmax(voltages)
            ]  # Argmax = Returns the indices of the maximum values along an axis.
            guess_offset = np.mean(
                voltages[np.abs(voltages - np.mean(voltages) < np.std(voltages))]
            )
            guess_sigma = abs(frequencies[np.argmin(voltages)] - guess_center)
            guess_amp = (np.max(voltages) - guess_offset) * guess_sigma * np.pi

        else:
            guess_center = frequencies[
                np.argmin(voltages)
            ]  # Argmin = Returns the indices of the minimum values along an axis.
            guess_offset = np.mean(
                voltages[np.abs(voltages - np.mean(voltages) < np.std(voltages))]
            )
            guess_sigma = abs(frequencies[np.argmax(voltages)] - guess_center)
            guess_amp = (np.min(voltages) - guess_offset) * guess_sigma * np.pi

        # Add guessed parameters to the model
        model_Q.set_param_hint("center", value=guess_center, vary=True)
        model_Q.set_param_hint("sigma", value=guess_sigma, vary=True)
        model_Q.set_param_hint("amplitude", value=guess_amp, vary=True)
        model_Q.set_param_hint("offset", value=guess_offset, vary=True)
        guess_parameters = model_Q.make_params()

        # fit the model with the data and guessed parameters
        try:
            fit_res = model_Q.fit(
                data=voltages, frequency=frequencies, params=guess_parameters
            )
        except:
            log.warning("lorentzian_fit: the fitting was not successful")
            data_fit.add({key: 0 for key in data_fit.df.columns})
            return data_fit

        # get the values for postprocessing and for legend.
        f0 = fit_res.best_values["center"]
        BW = fit_res.best_values["sigma"] * 2
        Q = abs(f0 / BW)
        peak_voltage = (
            fit_res.best_values["amplitude"] / (fit_res.best_values["sigma"] * np.pi)
            + fit_res.best_values["offset"]
        )

        freq = f0

        data_fit.add(
            {
                labels[0]: freq,
                labels[1]: peak_voltage,
                "popt0": fit_res.best_values["amplitude"],
                "popt1": fit_res.best_values["center"],
                "popt2": fit_res.best_values["sigma"],
                "popt3": fit_res.best_values["offset"],
                "qubit": qubit,
            }
        )
    return data_fit


def rabi_fit(data, x, y, qubits, resonator_type, labels):
    r"""
    Fitting routine for Rabi experiment. The used model is

    .. math::

        y = p_0 + p_1 sin(2 \pi p_2 x + p_3) e^{-x p_4}.

    Args:

        data (`DataUnits`): dataset for the fit
        x (str): name of the input values for the Rabi model
        y (str): name of the output values for the Rabi model
        qubits (list): A list with the IDs of the qubits
        resonator_type (str): the type of readout resonator ['3D', '2D']
        labels (list of str): list containing the lables of the quantities computed by this fitting method.

    Returns:

        A ``Data`` object with the following keys

            - **popt0**: offset
            - **popt1**: oscillation amplitude
            - **popt2**: frequency
            - **popt3**: phase
            - **popt4**: T2
            - **labels[0]**: pulse parameter
            - **labels[1]**: pulse's maximum voltage
            - **qubit**: The qubit being tested
    """

    data_fit = Data(
        name=f"fits",
        quantities=[
            "popt0",
            "popt1",
            "popt2",
            "popt3",
            "popt4",
            labels[0],
            labels[1],
            "qubit",
        ],
    )

    parameter_keys = parse(x)
    voltages_keys = parse(y)
    for qubit in qubits:
        qubit_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby(parameter_keys[0], as_index=False)
            .mean()
        )
        parameter = (
            qubit_data[parameter_keys[0]].pint.to(parameter_keys[1]).pint.magnitude
        )
        voltages = qubit_data[voltages_keys[0]].pint.to(voltages_keys[1]).pint.magnitude

        if resonator_type == "3D":
            pguess = [
                np.mean(voltages.values),
                np.max(voltages.values) - np.min(voltages.values),
                0.5 / parameter.values[np.argmin(voltages.values)],
                np.pi / 2,
                0.1e-6,
            ]
        else:
            pguess = [
                np.mean(voltages.values),
                np.max(voltages.values) - np.min(voltages.values),
                0.5 / parameter.values[np.argmax(voltages.values)],
                np.pi / 2,
                0.1e-6,
            ]
        try:
            popt, pcov = curve_fit(
                rabi, parameter.values, voltages.values, p0=pguess, maxfev=10000
            )
            smooth_dataset = rabi(parameter.values, *popt)
            pi_pulse_parameter = np.abs((1.0 / popt[2]) / 2)
            pi_pulse_peak_voltage = smooth_dataset.max()
            t2 = 1.0 / popt[4]  # double check T1
        except:
            log.warning("rabi_fit: the fitting was not succesful")
            data_fit.add({key: 0 for key in data_fit.df.columns})
            return data_fit

        data_fit.add(
            {
                "popt0": popt[0],
                "popt1": popt[1],
                "popt2": popt[2],
                "popt3": popt[3],
                "popt4": popt[4],
                labels[0]: pi_pulse_parameter,
                labels[1]: pi_pulse_peak_voltage,
                "qubit": qubit,
            }
        )
    return data_fit


def ramsey_fit(
    data, x, y, qubits, resonator_type, qubit_freqs, sampling_rate, offset_freq, labels
):
    r"""
    Fitting routine for Ramsey experiment. The used model is

    .. math::

        y = p_0 + p_1 sin \Big(p_2 x + p_3 \Big) e^{-x p_4}.

    Args:

        data (`DataUnits`): dataset for the fit
        x (str): name of the input values for the Ramsey model
        y (str): name of the output values for the Ramsey model
        qubits (list): A list with the IDs of the qubits
        qubits_freq (float): frequency of the qubit
        sampling_rate (float): Platform sampling rate
        offset_freq (float): Total qubit frequency offset. It contains the artificial detunning applied
                             by the experimentalist + the inherent offset in the actual qubit frequency stored in the runcard.
        labels (list of str): list containing the lables of the quantities computed by this fitting method.

    Returns:

        A ``Data`` object with the following keys

            - **popt0**: offset
            - **popt1**: oscillation amplitude
            - **popt2**: frequency
            - **popt3**: phase
            - **popt4**: T2
            - **labels[0]**: Physical detunning of the actual qubit frequency
            - **labels[1]**: New qubit frequency after correcting the actual qubit frequency with the detunning calculated (labels[0])
            - **labels[2]**: T2
            - **qubit**: The qubit being tested
    """
    data_fit = Data(
        name=f"fits",
        quantities=[
            "popt0",
            "popt1",
            "popt2",
            "popt3",
            "popt4",
            labels[0],
            labels[1],
            labels[2],
            "qubit",
        ],
    )

    parameter_keys = parse(x)
    voltages_keys = parse(y)
    for qubit in qubits:
        qubit_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby(parameter_keys[0], as_index=False)
            .mean()
        )
        times = qubit_data[parameter_keys[0]].pint.to(parameter_keys[1]).pint.magnitude
        voltages = qubit_data[voltages_keys[0]].pint.to(voltages_keys[1]).pint.magnitude

        try:
            y_max = np.max(voltages.values)
            y_min = np.min(voltages.values)
            y = (voltages.values - y_min) / (y_max - y_min)
            x_max = np.max(times.values)
            x_min = np.min(times.values)
            x = (times.values - x_min) / (x_max - x_min)
            if resonator_type == "3D":
                index = np.argmin(y)
            else:
                index = np.argmax(y)

            p0 = [
                np.mean(y),
                y_max - y_min,
                0.5 / x[index],
                np.pi / 2,
                0,
            ]
            popt = curve_fit(ramsey, x, y, method="lm", p0=p0)[0]
            popt = [
                (y_max - y_min) * popt[0] + y_min,
                (y_max - y_min) * popt[1] * np.exp(x_min * popt[4] / (x_max - x_min)),
                popt[2] / (x_max - x_min),
                popt[3] - x_min * popt[2] / (x_max - x_min),
                popt[4] / (x_max - x_min),
            ]
            delta_fitting = popt[2] / 2 * np.pi
            delta_phys = int((delta_fitting * sampling_rate) - offset_freq)
            corrected_qubit_frequency = int(qubit_freqs[qubit] + delta_phys)
            t2 = 1.0 / popt[4]
        except:
            log.warning("ramsey_fit: the fitting was not succesful")
            data_fit.add({key: 0 for key in data_fit.df.columns})
            return data_fit

        data_fit.add(
            {
                "popt0": popt[0],
                "popt1": popt[1],
                "popt2": popt[2],
                "popt3": popt[3],
                "popt4": popt[4],
                labels[0]: delta_phys,
                labels[1]: corrected_qubit_frequency,
                labels[2]: t2,
                "qubit": qubit,
            }
        )
    return data_fit


def t1_fit(data, x, y, qubits, resonator_type, labels):
    """
    Fitting routine for T1 experiment. The used model is

        .. math::

            y = p_0-p_1 e^{-x p_2}.

    Args:

        data (`DataUnits`): dataset for the fit
        x (str): name of the input values for the T1 model
        y (str): name of the output values for the T1 model
        qubit (int): ID qubit number
        nqubits (int): total number of qubits
        labels (list of str): list containing the lables of the quantities computed by this fitting method.

    Returns:

        A ``Data`` object with the following keys

            - **popt0**: p0
            - **popt1**: p1
            - **popt2**: p2
            - **labels[0]**: T1.

    """

    data_fit = Data(
        name=f"fits",
        quantities=[
            "popt0",
            "popt1",
            "popt2",
            labels[0],
        ],
    )

    parameter_keys = parse(x)
    voltages_keys = parse(y)
    for qubit in qubits:
        qubit_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby(parameter_keys[0], as_index=False)
            .mean()
        )
        times = qubit_data[parameter_keys[0]].pint.to(parameter_keys[1]).pint.magnitude
        voltages = qubit_data[voltages_keys[0]].pint.to(voltages_keys[1]).pint.magnitude

        if resonator_type == "3D":
            pguess = [
                max(voltages.values),
                (max(voltages.values) - min(voltages.values)),
                1 / 250,
            ]
        else:
            pguess = [
                min(voltages.values),
                (max(voltages.values) - min(voltages.values)),
                1 / 250,
            ]

        try:
            popt, pcov = curve_fit(
                exp, times.values, voltages.values, p0=pguess, maxfev=2000000
            )
            t1 = abs(1 / popt[2])

        except:
            log.warning("t1_fit: the fitting was not succesful")
            data_fit.add({key: 0 for key in data_fit.df.columns})
            return data_fit

    data_fit.add(
        {
            "popt0": popt[0],
            "popt1": popt[1],
            "popt2": popt[2],
            labels[0]: t1,
            "qubit": qubit,
        }
    )
    return data_fit


def flipping_fit(data, x, y, qubits, resonator_type, pi_pulse_amplitudes, labels):
    r"""
    Fitting routine for T1 experiment. The used model is

    .. math::

        y = p_0 sin\Big(\frac{2 \pi x}{p_2} + p_3\Big).

    Args:

        data (`DataUnits`): dataset for the fit
        x (str): name of the input values for the flipping model
        y (str): name of the output values for the flipping model
        qubit (int): ID qubit number
        nqubits (int): total number of qubits
        niter(int): Number of times of the flipping sequence applied to the qubit
        pi_pulse_amplitudes(list): list of corrected pi pulse amplitude
        labels (list of str): list containing the lables of the quantities computed by this fitting method.

    Returns:

        A ``Data`` object with the following keys

            - **popt0**: p0
            - **popt1**: p1
            - **popt2**: p2
            - **popt3**: p3
            - **labels[0]**: delta amplitude
            - **labels[1]**: corrected amplitude


    """
    data_fit = Data(
        name="fits",
        quantities=[
            "popt0",
            "popt1",
            "popt2",
            "popt3",
            labels[0],
            labels[1],
            "qubit",
        ],
    )

    for qubit in qubits:
        qubit_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby("flips", as_index=False)
            .mean()
        )
        flips_keys = parse(x)
        voltages_keys = parse(y)
        flips = qubit_data[flips_keys[0]].pint.to(flips_keys[1]).pint.magnitude
        voltages = qubit_data[voltages_keys[0]].pint.to(voltages_keys[1]).pint.magnitude

        if resonator_type == "3D":
            pguess = [0.0003, np.mean(voltages), -18, 0]  # epsilon guess parameter
        else:
            pguess = [0.0003, np.mean(voltages), 18, 0]  # epsilon guess parameter

        try:
            popt, pcov = curve_fit(flipping, flips, voltages, p0=pguess, maxfev=2000000)
            epsilon = -np.pi / popt[2]
            amplitude_correction_factor = np.pi / (np.pi + epsilon)
            corrected_amplitude = (
                amplitude_correction_factor * pi_pulse_amplitudes[qubit]
            )
            # angle = (niter * 2 * np.pi / popt[2] + popt[3]) / (1 + 4 * niter)
            # amplitude_delta = angle * 2 / np.pi * pi_pulse_amplitude
        except:
            log.warning("flipping_fit: the fitting was not succesful")
            data_fit.add({key: 0 for key in data_fit.df.columns})
            return data_fit

        data_fit.add(
            {
                "popt0": popt[0],
                "popt1": popt[1],
                "popt2": popt[2],
                "popt3": popt[3],
                labels[0]: amplitude_correction_factor,
                labels[1]: corrected_amplitude,
                "qubit": qubit,
            }
        )

    return data_fit


def drag_tuning_fit(data: Data, x, y, qubits, labels):
    r"""
    Fitting routine for drag tunning. The used model is

        .. math::

            y = p_1 cos \Big(\frac{2 \pi x}{p_2} + p_3 \Big) + p_0.

    Args:

        data (`DataUnits`): dataset for the fit
        x (str): name of the input values for the model
        y (str): name of the output values for the model
        qubit (int): ID qubit number
        labels (list of str): list containing the lables of the quantities computed by this fitting method.

    Returns:

        A ``Data`` object with the following keys

            - **popt0**: offset
            - **popt1**: oscillation amplitude
            - **popt2**: period
            - **popt3**: phase
            - **labels[0]**: optimal beta.


    """

    data_fit = Data(
        name=f"fits",
        quantities=[
            "popt0",
            "popt1",
            "popt2",
            "popt3",
            labels[0],
            "qubit",
        ],
    )

    for qubit in qubits:
        qubit_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby("beta_param", as_index=False)
            .mean()
        )
        beta_params_keys = parse(x)
        voltages_keys = parse(y)
        beta_params = (
            qubit_data[beta_params_keys[0]].pint.to(beta_params_keys[1]).pint.magnitude
        )
        voltages = qubit_data[voltages_keys[0]].pint.to(voltages_keys[1]).pint.magnitude

        pguess = [
            0,  # Offset:    p[0]
            beta_params.values[np.argmax(voltages)]
            - beta_params.values[np.argmin(voltages)],  # Amplitude: p[1]
            4,  # Period:    p[2]
            0.3,  # Phase:     p[3]
        ]

        try:
            popt, pcov = curve_fit(cos, beta_params.values, voltages.values)
            smooth_dataset = cos(beta_params.values, popt[0], popt[1], popt[2], popt[3])
            beta_optimal = beta_params.values[np.argmin(smooth_dataset)]

        except:
            log.warning("drag_tuning_fit: the fitting was not succesful")
            data_fit.add({key: 0 for key in data_fit.df.columns})
            return data_fit

        data_fit.add(
            {
                "popt0": popt[0],
                "popt1": popt[1],
                "popt2": popt[2],
                "popt3": popt[3],
                labels[0]: beta_optimal,
                "qubit": qubit,
            }
        )
    return data_fit


def res_spectroscopy_flux_fit(data, x, y, qubit, fluxline, params_fit):
    """Fit frequency as a function of current for the flux resonator spectroscopy
        Args:
        data (DataUnits): Data file with information on the feature response at each current point.
        x (str): Column of the data file associated to x-axis.
        y (str): Column of the data file associated to y-axis.
        qubit (int): qubit coupled to the resonator that we are probing.
        fluxline (int): id of the current line used for the experiment.
        params_fit (list): List of parameters for the fit. [freq_rh, g, Ec, Ej].
                          freq_rh is the resonator frequency at high power and g in the readout coupling.
                          If Ec and Ej are missing, the fit is valid in the transmon limit and if they are indicated,
                          contains the next-order correction.

    Returns:
        data_fit (Data): Data file with labels and fit parameters.

    """

    curr = np.array(data.get_values(*parse(x)))
    freq = np.array(data.get_values(*parse(y)))
    if qubit == fluxline:
        if len(params_fit) == 2:
            quantities = [
                "curr_sp",
                "xi",
                "d",
                "f_q/f_rh",
                "g",
                "f_rh",
                "f_qs",
                "f_rs",
                "f_offset",
                "C_ii",
            ]
        else:
            quantities = [
                "curr_sp",
                "xi",
                "d",
                "g",
                "Ec",
                "Ej",
                "f_rh",
                "f_qs",
                "f_rs",
                "f_offset",
                "C_ii",
            ]

        data_fit = Data(
            name=f"fit1_q{qubit}_f{fluxline}",
            quantities=quantities,
        )
        try:
            f_rh = params_fit[0]
            g = params_fit[1]
            max_c = curr[np.argmax(freq)]
            min_c = curr[np.argmin(freq)]
            xi = 1 / (2 * abs(max_c - min_c))
            if len(params_fit) == 2:
                f_r = np.max(freq)
                f_q_0 = f_rh - g**2 / (f_r - f_rh)
                popt = curve_fit(
                    freq_r_transmon,
                    curr,
                    freq,
                    p0=[max_c, xi, 0, f_q_0 / f_rh, g, f_rh],
                )[0]
                f_qs = popt[3] * popt[5]
                f_rs = freq_r_transmon(popt[0], *popt)
                f_offset = freq_r_transmon(0, *popt)
                C_ii = (f_rs - f_offset) / popt[0]
                data_fit.add(
                    {
                        "curr_sp": popt[0],
                        "xi": popt[1],
                        "d": abs(popt[2]),
                        "f_q/f_rh": popt[3],
                        "g": popt[4],
                        "f_rh": popt[5],
                        "f_qs": f_qs,
                        "f_rs": f_rs,
                        "f_offset": f_offset,
                        "C_ii": C_ii,
                    }
                )
            else:
                Ec = params_fit[2]
                Ej = params_fit[3]
                freq_r_mathieu1 = partial(freq_r_mathieu, p7=0.4999)
                popt = curve_fit(
                    freq_r_mathieu1,
                    curr,
                    freq,
                    p0=[f_rh, g, max_c, xi, 0, Ec, Ej],
                    method="dogbox",
                )[0]
                f_qs = freq_q_mathieu(popt[2], *popt[2::])
                f_rs = freq_r_mathieu(popt[2], *popt)
                f_offset = freq_r_mathieu(0, *popt)
                C_ii = (f_rs - f_offset) / popt[2]
                data_fit.add(
                    {
                        "curr_sp": popt[2],
                        "xi": popt[3],
                        "d": abs(popt[4]),
                        "g": popt[1],
                        "Ec": popt[5],
                        "Ej": popt[6],
                        "f_rh": popt[0],
                        "f_qs": f_qs,
                        "f_rs": f_rs,
                        "f_offset": f_offset,
                        "C_ii": C_ii,
                    }
                )
        except:
            log.warning("The fitting was not successful")
            return data_fit
    else:
        data_fit = Data(
            name=f"fit1_q{qubit}_f{fluxline}",
            quantities=[
                "popt0",
                "popt1",
            ],
        )
        try:
            freq_min = np.min(freq)
            freq_max = np.max(freq)
            freq_norm = (freq - freq_min) / (freq_max - freq_min)
            popt = curve_fit(line, curr, freq_norm)[0]
            popt[0] = popt[0] * (freq_max - freq_min)
            popt[1] = popt[1] * (freq_max - freq_min) + freq_min
        except:
            log.warning("The fitting was not successful")
            return data_fit

        data_fit.add(
            {
                "popt0": popt[0],  # C_ij
                "popt1": popt[1],
            }
        )
    return data_fit


def res_spectroscopy_flux_matrix(folder, fluxlines):
    """Calculation of the resonator flux matrix, Mf.
       curr = Mf*freq + offset_c.
       Mf = Mc^-1, offset_c = -Mc^-1 * offset_f
       freq = Mc*curr + offset_f
        Args:
        folder (str): Folder where the data files with the experimental and fit data are.
        fluxlines (list): ids of the current line used for the experiment.

    Returns:
        data (Data): Data file with len(fluxlines)+1 columns that contains the flux matrix (Mf) and
                     offset (offset_c) in the last column.

    """
    import os

    from pandas import DataFrame

    fits = []
    for q in fluxlines:
        for f in fluxlines:
            file = f"{folder}/data/resonator_flux_sample/fit1_q{q}_f{f}.csv"
            if os.path.exists(file):
                fits += [f]
    if len(fits) == len(fluxlines) ** 2:
        mat = np.zeros((len(fluxlines), len(fluxlines)))
        offset = np.zeros(len(fluxlines))
        for i, q in enumerate(fluxlines):
            for j, f in enumerate(fluxlines):
                data_fit = Data.load_data(
                    folder, "data", "resonator_flux_sample", "csv", f"fit1_q{q}_f{f}"
                )
                if q == f:
                    element = "C_ii"
                    offset[i] = data_fit.get_values("f_offset")[0]
                else:
                    element = "popt0"
                mat[i, j] = data_fit.get_values(element)[0]
        m = np.linalg.inv(mat)
        offset_c = -m @ offset
        data = Data(name=f"flux_matrix")
        data.df = DataFrame(m)
        data.df.insert(len(fluxlines), "offset_c", offset_c, True)
        # [m, offset_c] freq = M*curr + offset --> curr = m*freq + offset_c  m = M^-1, offset_c = -M^-1 * offset
        data.to_csv(f"{folder}/data/resonator_flux_sample/")
    else:
        data = Data(name=f"flux_matrix")
    return data


def spin_echo_fit(data, x, y, qubits, resonator_type, labels):
    data_fit = Data(
        name=f"fits",
        quantities=[
            "popt0",
            "popt1",
            "popt2",
            labels[0],
        ],
    )

    parameter_keys = parse(x)
    voltages_keys = parse(y)
    for qubit in qubits:
        qubit_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby(parameter_keys[0], as_index=False)
            .mean()
        )
        times = qubit_data[parameter_keys[0]].pint.to(parameter_keys[1]).pint.magnitude
        voltages = qubit_data[voltages_keys[0]].pint.to(voltages_keys[1]).pint.magnitude

        if resonator_type == "3D":
            pguess = [
                max(voltages.values),
                (max(voltages.values) - min(voltages.values)),
                1 / 250,
            ]
        else:
            pguess = [
                min(voltages.values),
                (max(voltages.values) - min(voltages.values)),
                1 / 250,
            ]

        try:
            popt, pcov = curve_fit(
                exp, times.values, voltages.values, p0=pguess, maxfev=2000000
            )
            t2 = abs(1 / popt[2])

        except:
            log.warning("spin_echo_fit: the fitting was not succesful")
            return data_fit

    data_fit.add(
        {
            "popt0": popt[0],
            "popt1": popt[1],
            "popt2": popt[2],
            labels[0]: t2,
            "qubit": qubit,
        }
    )
    return data_fit


def calibrate_qubit_states_fit(data, x, y, nshots, qubits):
    parameters = Data(
        name=f"parameters",
        quantities={
            "rotation_angle",  # in degrees
            "threshold",
            "fidelity",
            "assignment_fidelity",
            "average_state0",
            "average_state1",
            "qubit",
        },
    )

    i_keys = parse(x)
    q_keys = parse(y)

    for qubit in qubits:
        qubit_data = data.df[data.df["qubit"] == qubit]

        iq_state0 = (
            qubit_data[qubit_data["state"] == 0][i_keys[0]]
            .pint.to(i_keys[1])
            .pint.magnitude
            + 1.0j
            * qubit_data[qubit_data["state"] == 0][q_keys[0]]
            .pint.to(q_keys[1])
            .pint.magnitude
        )
        iq_state1 = (
            qubit_data[qubit_data["state"] == 1][i_keys[0]]
            .pint.to(i_keys[1])
            .pint.magnitude
            + 1.0j
            * qubit_data[qubit_data["state"] == 1][q_keys[0]]
            .pint.to(q_keys[1])
            .pint.magnitude
        )

    for qubit in qubits:
        qubit_data = data.df[data.df["qubit"] == qubit]

        iq_state0 = (
            qubit_data[qubit_data["state"] == 0][i_keys[0]]
            .pint.to(i_keys[1])
            .pint.magnitude
            + 1.0j
            * qubit_data[qubit_data["state"] == 0][q_keys[0]]
            .pint.to(q_keys[1])
            .pint.magnitude
        )
        iq_state1 = (
            qubit_data[qubit_data["state"] == 1][i_keys[0]]
            .pint.to(i_keys[1])
            .pint.magnitude
            + 1.0j
            * qubit_data[qubit_data["state"] == 1][q_keys[0]]
            .pint.to(q_keys[1])
            .pint.magnitude
        )

        iq_state1 = np.array(iq_state1)
        iq_state0 = np.array(iq_state0)

        iq_mean_state1 = np.mean(iq_state1)
        iq_mean_state0 = np.mean(iq_state0)
        origin = iq_mean_state0

        iq_state0_translated = iq_state0 - origin
        iq_state1_translated = iq_state1 - origin
        rotation_angle = np.angle(np.mean(iq_state1_translated))

        iq_state1_rotated = iq_state1 * np.exp(-1j * rotation_angle)
        iq_state0_rotated = iq_state0 * np.exp(-1j * rotation_angle)

        real_values_state1 = iq_state1_rotated.real
        real_values_state0 = iq_state0_rotated.real

        real_values_combined = np.concatenate((real_values_state1, real_values_state0))
        real_values_combined.sort()

        cum_distribution_state1 = [
            sum(map(lambda x: x.real >= real_value, real_values_state1))
            for real_value in real_values_combined
        ]
        cum_distribution_state0 = [
            sum(map(lambda x: x.real >= real_value, real_values_state0))
            for real_value in real_values_combined
        ]

        cum_distribution_diff = np.abs(
            np.array(cum_distribution_state1) - np.array(cum_distribution_state0)
        )
        argmax = np.argmax(cum_distribution_diff)
        threshold = real_values_combined[argmax]
        errors_state1 = nshots - cum_distribution_state1[argmax]
        errors_state0 = cum_distribution_state0[argmax]
        fidelity = cum_distribution_diff[argmax] / nshots
        assignment_fidelity = 1 - (errors_state1 + errors_state0) / nshots / 2
        # assignment_fidelity = 1/2 + (cum_distribution_state1[argmax] - cum_distribution_state0[argmax])/nshots/2

        results = {
            "rotation_angle": (-rotation_angle * 360 / (2 * np.pi)) % 360,  # in degrees
            "threshold": threshold,
            "fidelity": fidelity,
            "assignment_fidelity": assignment_fidelity,
            "average_state0": iq_mean_state0,
            "average_state1": iq_mean_state1,
            "qubit": qubit,
        }
        parameters.add(results)
    return parameters



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal import plots
from qibocal.config import raise_error
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import lorentzian_fit


@plot("MSR and Phase vs Qubit Drive Frequency", plots.frequency_msr_phase)
def qubit_spectroscopy(
    platform: AbstractPlatform,
    qubits: dict,
    fast_width,
    fast_step,
    precision_width,
    precision_step,
    software_averages=1,
    points=10,
):
    r"""
    Perform spectroscopy on the qubit.
    This routine executes a fast scan around the expected qubit frequency indicated in the platform runcard.
    Afterthat, a final sweep with more precision is executed centered in the new qubit frequency found.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        fast_start (int): Initial frequency in HZ to perform the qubit fast sweep
        fast_width (int): Width frequency in HZ to perform the high resolution sweep
        fast_step (int): Step frequency in HZ for the high resolution sweep
        precision_width (int): Width frequency in HZ to perform the precision resolution sweep
        precision_step (int): Step frequency in HZ for the precission resolution sweep
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - Two DataUnits objects with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Qubit drive frequency value in Hz
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **qubit**: The qubit being tested
            - **drive_frequency**: frequency
            - **peak_voltage**: peak voltage
            - **popt0**: Lorentzian's amplitude
            - **popt1**: Lorentzian's center
            - **popt2**: Lorentzian's sigma
            - **popt3**: Lorentzian's offset
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment:
    # long drive probing pulse - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(
            qubit, start=0, duration=5000
        )
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=5000)
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    delta_frequency_range = np.arange(-fast_width // 2, fast_width // 2, fast_step)

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit frequency
    fast_sweep_data = DataUnits(
        name="fast_sweep_data",
        quantities={"frequency": "Hz"},
        options=["qubit", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        # sweep the parameter
        for delta_freq in delta_frequency_range:
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield fast_sweep_data
                # calculate and save fit
                yield lorentzian_fit(
                    fast_sweep_data,
                    x="frequency[Hz]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=["drive_frequency", "peak_voltage"],
                )
            # reconfigure the instruments based on the new resonator frequency
            # in this case setting the local oscillators
            # the pulse sequence does not need to be modified or recreated between executions
            for qubit in qubits:
                qd_pulses[qubit].frequency = delta_freq + qubits[qubit].drive_frequency

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            # retrieve the results for every qubit
            for ro_pulse in ro_pulses.values():
                # average msr, phase, i and q over the number of shots defined in the runcard
                r = results[ro_pulse.serial].to_dict()
                # store the results
                r.update(
                    {
                        "frequency[Hz]": qd_pulses[ro_pulse.qubit].frequency,
                        "qubit": ro_pulse.qubit,
                        "iteration": iteration,
                    }
                )
                fast_sweep_data.add(r)
            count += 1
    # finally, save the remaining data and fits
    yield fast_sweep_data
    yield lorentzian_fit(
        fast_sweep_data,
        x="frequency[Hz]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=["drive_frequency", "peak_voltage"],
    )

    # store max/min peaks as new frequencies
    for qubit in qubits:
        qubit_data = (
            fast_sweep_data.df[fast_sweep_data.df["qubit"] == qubit]
            .drop(columns=["qubit", "iteration"])
            .groupby("frequency", as_index=False)
            .mean()
        )
        if platform.resonator_type == "3D":
            qubits[qubit].drive_frequency = (
                qubit_data["frequency"][
                    np.argmin(qubit_data["MSR"].pint.to("V").pint.magnitude)
                ]
                .to("Hz")
                .magnitude
            )
        else:
            qubits[qubit].drive_frequency = (
                qubit_data["frequency"][
                    np.argmax(qubit_data["MSR"].pint.to("V").pint.magnitude)
                ]
                .to("Hz")
                .magnitude
            )

    # run a precision sweep around the newly detected frequencies

    delta_frequency_range = np.arange(
        -precision_width // 2, precision_width // 2, precision_step
    )

    # create a second DataUnits object to store the results,
    precision_sweep_data = DataUnits(
        name="precision_sweep_data",
        quantities={"frequency": "Hz"},
        options=["qubit", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        # sweep the parameter
        for delta_freq in delta_frequency_range:
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield precision_sweep_data
                # calculate and save fit
                yield lorentzian_fit(
                    precision_sweep_data,
                    x="frequency[Hz]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=["resonator_freq", "peak_voltage"],
                )
            # reconfigure the instrument based on the new resonator frequency
            # in this case setting the local oscillators
            # the pulse sequence does not need to be modified between executions
            for qubit in qubits:
                qd_pulses[qubit].frequency = delta_freq + qubits[qubit].drive_frequency

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            # retrieve the results for every qubit
            for ro_pulse in ro_pulses.values():
                # average msr, phase, i and q over the number of shots defined in the runcard
                r = results[ro_pulse.serial].to_dict()
                # store the results
                r.update(
                    {
                        "frequency[Hz]": qd_pulses[ro_pulse.qubit].frequency,
                        "qubit": ro_pulse.qubit,
                        "iteration": iteration,
                    }
                )
                precision_sweep_data.add(r)
            count += 1
    # finally, save the remaining data and fits
    yield precision_sweep_data
    yield lorentzian_fit(
        precision_sweep_data,
        x="frequency[Hz]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=["drive_frequency", "peak_voltage"],
    )


@plot(
    "MSR and Phase vs Qubit Drive Frequency and Flux Current",
    plots.frequency_flux_msr_phase,
)
def qubit_spectroscopy_flux(
    platform: AbstractPlatform,
    qubits: dict,
    freq_width,
    freq_step,
    bias_width,
    bias_step,
    fluxlines,
    software_averages=1,
    points=10,
):
    r"""
    Perform spectroscopy on the qubit modifying the bias applied in the flux control line.
    This routine works for multiqubit devices flux controlled.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        freq_width (int): Width frequency in HZ to perform the spectroscopy sweep
        freq_step (int): Step frequency in HZ for the spectroscopy sweep
        bias_width (float): Width bias in A for the flux bias sweep
        bias_step (float): Step bias in A for the flux bias sweep
        fluxlines (list): List of flux lines to use to perform the experiment. If it is set to "qubits", it uses each of
                        flux lines associated with the target qubits.
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Qubit drive frequency value in Hz
            - **bias[V]**: Current value in A applied to the flux line
            - **qubit**: The qubit being tested
            - **fluxline**: The fluxline being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **drive_frequency**: frequency
            - **peak_voltage**: peak voltage
            - **popt0**: Lorentzian's amplitude
            - **popt1**: Lorentzian's center
            - **popt2**: Lorentzian's sigma
            - **popt3**: Lorentzian's offset
            - **qubit**: The qubit being tested
    """
    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment:
    # long drive probing pulse - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(
            qubit, start=0, duration=5000
        )
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=5000)
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # qubit drive frequency
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)

    # flux bias
    sweetspot_biass = {}
    bias_ranges = {}
    bias_min = {}
    bias_max = {}

    if fluxlines == "qubits":
        fluxlines = qubits

    for fluxline in fluxlines:
        # TODO: check if this is correct
        sweetspot_biass[fluxline] = qubits[fluxline].sweetspot

        bias_min[fluxline] = max(-bias_width / 2 + sweetspot_biass[fluxline], -0.03)
        bias_max[fluxline] = min(+bias_width / 2 + sweetspot_biass[fluxline], +0.03)
        bias_ranges[fluxline] = np.arange(
            bias_min[fluxline], bias_max[fluxline], bias_step
        )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit frequency and flux bias
    data = DataUnits(
        name=f"data",
        quantities={"frequency": "Hz", "bias": "V"},
        options=["qubit", "fluxline", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        # sweep the parameters
        for fluxline in fluxlines:
            for bias in bias_ranges[fluxline]:
                # set new flux bias
                platform.set_bias(fluxline, bias)

                # TODO: adjust resonator frequency if coefs available in the runcard
                # coefs should be determined in resonator_spectroscopy_flux
                # matrix of biass -> magnetic flux -> freq shift

                for delta_freq in delta_frequency_range:
                    # save data as often as defined by points
                    if count % points == 0:
                        # save data
                        yield data
                        # TODO: calculate and save fit

                    # set new lo frequency
                    for qubit in qubits:
                        qd_pulses[qubit].frequency = (
                            delta_freq + qubits[qubit].drive_frequency
                        )

                    # execute the pulse sequence
                    result = platform.execute_pulse_sequence(sequence)

                    # retrieve the results for every qubit
                    for ro_pulse in ro_pulses.values():
                        # average msr, phase, i and q over the number of shots defined in the runcard
                        r = result[ro_pulse.serial].to_dict()
                        # store the results
                        r.update(
                            {
                                "frequency[Hz]": qd_pulses[ro_pulse.qubit].frequency,
                                "bias[V]": bias,
                                "qubit": ro_pulse.qubit,
                                "fluxline": fluxline,
                                "iteration": iteration,
                            }
                        )
                        data.add(r)
                    count += 1
    # finally, save the remaining data and fits
    yield data



import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.data import Data, DataUnits
from qibocal.fitting.utils import cos
from qibocal.plots.utils import get_color, get_data_subfolders

# allXY rotations
gatelist = [
    ["I", "I"],
    ["RX(pi)", "RX(pi)"],
    ["RY(pi)", "RY(pi)"],
    ["RX(pi)", "RY(pi)"],
    ["RY(pi)", "RX(pi)"],
    ["RX(pi/2)", "I"],
    ["RY(pi/2)", "I"],
    ["RX(pi/2)", "RY(pi/2)"],
    ["RY(pi/2)", "RX(pi/2)"],
    ["RX(pi/2)", "RY(pi)"],
    ["RY(pi/2)", "RX(pi)"],
    ["RX(pi)", "RY(pi/2)"],
    ["RY(pi)", "RX(pi/2)"],
    ["RX(pi/2)", "RX(pi)"],
    ["RX(pi)", "RX(pi/2)"],
    ["RY(pi/2)", "RY(pi)"],
    ["RY(pi)", "RY(pi/2)"],
    ["RX(pi)", "I"],
    ["RY(pi)", "I"],
    ["RX(pi/2)", "RX(pi/2)"],
    ["RY(pi/2)", "RY(pi/2)"],
]


def allXY(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data: -<br>No fitting data: -<br><br>"

    fig = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(f"allXY",),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    for subfolder in subfolders:
        try:
            data = Data.load_data(folder, subfolder, routine, format, "data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = Data(
                name="data",
                quantities={"probability", "gateNumber", "qubit", "iteration"},
            )
        iterations = data.df["iteration"].unique()
        gate_numbers = data.df["gateNumber"].unique()

        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["gateNumber"],
                    y=iteration_data["probability"],
                    marker_color=get_color(report_n),
                    mode="markers",
                    text=gatelist,
                    textposition="bottom center",
                    opacity=0.3,
                    name=f"q{qubit}/r{report_n}: Probability",
                    showlegend=not bool(iteration),
                    legendgroup="group1",
                ),
                row=1,
                col=1,
            )

        fig.add_trace(
            go.Scatter(
                x=gate_numbers,
                y=data.df.groupby("gateNumber", as_index=False)[
                    "probability"
                ].mean(),  # pylint: disable=E1101
                name=f"q{qubit}/r{report_n}: Average Probability",
                marker_color=get_color(report_n),
                mode="markers",
                text=gatelist,
                textposition="bottom center",
            ),
            row=1,
            col=1,
        )
        report_n += 1

    fig.add_hline(
        y=0,
        line_width=2,
        line_dash="dash",
        line_color="grey",
        row=1,
        col=1,
    )
    fig.add_hline(
        y=1,
        line_width=2,
        line_dash="dash",
        line_color="grey",
        row=1,
        col=1,
    )

    fig.add_hline(
        y=-1,
        line_width=2,
        line_dash="dash",
        line_color="grey",
        row=1,
        col=1,
    )

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Gate sequence number",
        yaxis_title="Expectation value of Z",
    )

    figures.append(fig)

    return figures, fitting_report


# allXY
def allXY_drag_pulse_tuning(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data: -<br>No fitting data: -<br><br>"

    fig = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(f"allXY",),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    for subfolder in subfolders:
        try:
            data = Data.load_data(folder, subfolder, routine, format, "data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = Data(
                quantities={
                    "probability",
                    "gateNumber",
                    "beta_param",
                    "qubit",
                    "iteration",
                },
            )

        iterations = data.df["iteration"].unique()
        beta_params = data.df["beta_param"].unique()
        gate_numbers = data.df["gateNumber"].unique()

        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            for j, beta_param in enumerate(beta_params):
                beta_param_data = iteration_data[
                    iteration_data["beta_param"] == beta_param
                ]
                fig.add_trace(
                    go.Scatter(
                        x=beta_param_data["gateNumber"],
                        y=beta_param_data["probability"],
                        marker_color=get_color(report_n * len(beta_params) + j),
                        mode="markers+lines",
                        opacity=0.5,
                        name=f"q{qubit}/r{report_n}: beta {beta_param}",
                        showlegend=not bool(iteration),
                        legendgroup=f"group{j}",
                        text=gatelist,
                        textposition="bottom center",
                    ),
                    row=1,
                    col=1,
                )
        report_n += 1

    fig.add_hline(
        y=0,
        line_width=2,
        line_dash="dash",
        line_color="grey",
        row=1,
        col=1,
    )
    fig.add_hline(
        y=1,
        line_width=2,
        line_dash="dash",
        line_color="grey",
        row=1,
        col=1,
    )

    fig.add_hline(
        y=-1,
        line_width=2,
        line_dash="dash",
        line_color="grey",
        row=1,
        col=1,
    )

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Gate sequence number",
        yaxis_title="Expectation value of Z",
    )

    figures.append(fig)

    return figures, fitting_report


# beta param tuning
def drag_pulse_tuning(folder, routine, qubit, format):
    figures = []
    fitting_report = "No fitting data: -<br>No fitting data: -<br><br>"

    fig = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.01,
        vertical_spacing=0.01,
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, "data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name="data",
                quantities={"beta_param": "dimensionless"},
                options=["qubit", "iteration"],
            )
        try:
            data_fit = Data.load_data(folder, subfolder, routine, format, "fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except:
            data_fit = Data()

        iterations = data.df["iteration"].unique()
        beta_params = data.df["beta_param"].pint.magnitude.unique()

        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["beta_param"].pint.magnitude,
                    y=iteration_data["MSR"].pint.to("uV").pint.magnitude,
                    marker_color=get_color(report_n),
                    mode="markers",
                    opacity=0.3,
                    name=f"q{qubit}/r{report_n}: Probability",
                    showlegend=not bool(iteration),
                    legendgroup="group1",
                ),
                row=1,
                col=1,
            )

        fig.add_trace(
            go.Scatter(
                x=beta_params,
                y=data.df.groupby("beta_param", as_index=False)
                .mean()["MSR"]  # pylint: disable=E1101
                .pint.to("uV")
                .pint.magnitude,
                name=f"q{qubit}/r{report_n}: Average MSR",
                marker_color=get_color(report_n),
                mode="markers",
            ),
            row=1,
            col=1,
        )

        # add fitting traces

        if len(data) > 0 and (qubit in data_fit.df["qubit"].values):
            beta_range = np.linspace(
                min(data.get_values("beta_param", "dimensionless")),
                max(data.get_values("beta_param", "dimensionless")),
                20,
            )

            params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                orient="records"
            )[0]
            fig.add_trace(
                go.Scatter(
                    x=beta_range,
                    y=cos(
                        beta_range,
                        data_fit.get_values("popt0"),
                        data_fit.get_values("popt1"),
                        data_fit.get_values("popt2"),
                        data_fit.get_values("popt3"),
                    ),
                    name=f"q{qubit}/r{report_n}: Fit",
                    line=go.scatter.Line(dash="dot"),
                    marker_color=get_color(4 * report_n + 2),
                ),
                row=1,
                col=1,
            )
            fitting_report = fitting_report + (
                f"q{qubit}/r{report_n} optimal_beta_param: {params['optimal_beta_param']:.4f}<br><br>"
            )

            report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Beta parameter",
        yaxis_title="MSR[uV] [Rx(pi/2) - Ry(pi)] - [Ry(pi/2) - Rx(pi)]",
    )

    figures.append(fig)

    return figures, fitting_report



import datetime
import os
import re
import time

import pandas as pd
import yaml
from dash import Dash, Input, Output, dcc, html

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.web.server import server

DataUnits()  # dummy dataset call to suppress ``pint[V]`` error

app = Dash(
    server=server,
    suppress_callback_exceptions=True,
)

app.layout = html.Div(
    [
        html.Link(href="/static/styles.css", rel="stylesheet"),
        html.Div(
            id="refresh-area",
            className="refresh-area",
            children=[
                html.Div(
                    id="refresh-text",
                    className="refresh-text",
                    children=[html.P("Refresh rate:")],
                ),
                dcc.Dropdown(
                    id="interval-refresh",
                    className="interval-refresh",
                    placeholder="Select refresh rate",
                    options=[
                        {"label": "Auto refresh", "value": 0},
                        {"label": "2 seconds", "value": 2},
                        {"label": "5 seconds", "value": 5},
                        {"label": "10 seconds", "value": 10},
                        {"label": "20 seconds", "value": 20},
                        {"label": "No refresh", "value": 3600},
                    ],
                    value=0,
                ),
                html.Div(
                    id="latest-timestamp",
                    className="latest-timestamp",
                ),
            ],
        ),
        html.Div(id="div-fitting", className="div-fitting"),
        html.Div(id="div-figures"),
        dcc.Location(id="url", refresh=False),
        dcc.Interval(
            id="interval",
            # TODO: Perhaps the user should be allowed to change the refresh rate
            interval=5 * 1000,
            n_intervals=0,
            disabled=False,
        ),
    ]
)


@app.callback(
    Output(component_id="div-figures", component_property="children"),
    Output(component_id="latest-timestamp", component_property="children"),
    Output(component_id="interval", component_property="interval"),
    Output(component_id="div-fitting", component_property="children"),
    Input("interval", "n_intervals"),
    Input("url", "pathname"),
    Input("interval-refresh", "value"),
)
def get_graph(interval, url, value):
    st = time.time()

    figures = []

    if "data" not in url:
        url = f"/data{url}"

    method, folder, routine, qubit, format = url.split(os.sep)[2:]
    if qubit.isdigit():
        qubit = int(qubit)

    try:
        # data = DataUnits.load_data(folder, routine, format, "precision_sweep")
        # with open(f"{folder}/platform.yml", "r") as f:
        #     nqubits = yaml.safe_load(f)["nqubits"]
        # if len(data) > 2:
        #     params, fit = resonator_spectroscopy_fit(folder, format, nqubits)
        # else:
        #     params, fit = None, None
        # return getattr(plots.resonator_spectroscopy, method)(data, params, fit)

        # # FIXME: Temporarily hardcode the plotting method to test
        # # multiple routines with different names in one folder
        # # should be changed to:
        # # return getattr(getattr(plots, routine), method)(data)
        figs, fitting_report = getattr(plots, method)(folder, routine, qubit, format)
        et = time.time()

        if value == 0:
            refresh_rate = (et - st) + 6
        else:
            refresh_rate = value

        for fig in figs:
            figures.append(dcc.Graph(figure=fig))

        table = ""
        if "No fitting data" not in fitting_report:
            fitting_params = re.split(r"<br>|:|\|", fitting_report)
            fitting_params = list(filter(lambda x: x.strip(), fitting_params))
            table_header = [
                html.Thead(
                    children=[
                        html.Tr(
                            children=[
                                html.Th(
                                    className="th_styles", children="qubit # / report #"
                                ),
                                html.Th(
                                    className="th_styles", children="Fitting Parameter"
                                ),
                                html.Th(className="th_styles", children="Value"),
                            ]
                        )
                    ]
                )
            ]
            table_rows = []

            for i in range(0, len(fitting_params), 3):
                table_rows.append(
                    html.Tr(
                        className="td_styles",
                        children=[
                            html.Td(fitting_params[i]),
                            html.Td(fitting_params[i + 1]),
                            html.Td(fitting_params[i + 2]),
                        ],
                    )
                )

            table = [
                html.Table(
                    className="fitting-table", children=table_header + table_rows
                )
            ]

        timestamp = datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")
        return (
            figures,
            [html.Span(f"Last update: {(timestamp)}")],
            refresh_rate * 1000,
            table,
        )
    except (FileNotFoundError, pd.errors.EmptyDataError):
        timestamp = datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")
        return (
            figures,
            [html.Span(f"Last updated: {timestamp}")],
            refresh_rate * 1000,
            table,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['qibocal',
 'qibocal.calibrations',
 'qibocal.calibrations.characterization',
 'qibocal.calibrations.niGSC',
 'qibocal.calibrations.niGSC.basics',
 'qibocal.cli',
 'qibocal.fitting',
 'qibocal.plots',
 'qibocal.tests',
 'qibocal.web']

package_data = \
{'': ['*'], 'qibocal.web': ['static/*', 'templates/*']}

install_requires = \
['Pint-Pandas>=0.3,<0.4',
 'dash>=2.6.0,<3.0.0',
 'lmfit>=1.0.3,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'qibo>=0.1.9,<0.2.0',
 'qibolab>=0.0.1,<0.0.2']

extras_require = \
{'docs': ['Sphinx>=4.3.2,<5.0.0',
          'furo>=2022.9.29,<2023.0.0',
          'sphinxcontrib-bibtex>=2.4.1,<3.0.0',
          'recommonmark>=0.7.1,<0.8.0',
          'sphinx_markdown_tables>=0.0.17,<0.0.18']}

entry_points = \
{'console_scripts': ['qq = qibocal:command',
                     'qq-compare = qibocal:compare',
                     'qq-live = qibocal:live_plot',
                     'qq-upload = qibocal:upload']}

setup_kwargs = {
    'name': 'qibocal',
    'version': '0.0.1',
    'description': '',
    'long_description': '# Qibocal\n![Tests](https://github.com/qiboteam/qibocal/workflows/Tests/badge.svg)\n[![codecov](https://codecov.io/gh/qiboteam/qibocal/branch/main/graph/badge.svg?token=1EKZKVEVX0)](https://codecov.io/gh/qiboteam/qibo)\n[![Documentation Status](https://readthedocs.org/projects/qibocal/badge/?version=latest)](https://qibocal.readthedocs.io/en/latest/)\n\nQibocal provides Quantum Characterization Validation and Verification protocols using [Qibo](https://github.com/qiboteam/qibo) and [Qibolab](https://github.com/qiboteam/qibolab).\n\nQibocal key features:\n\n- Automatization of calibration routines.\n\n- Declarative inputs using runcard.\n\n- Generation of a report.\n\n## Installation\n\nThe package can be installed by source:\n```sh\ngit clone https://github.com/qiboteam/qibocal.git\ncd qibocal\npip install .\n```\n\n\n### Developer instructions\nFor development make sure to install the package using [`poetry`](https://python-poetry.org/) and to install the pre-commit hooks:\n```sh\ngit clone https://github.com/qiboteam/qibocal.git\ncd qibocal\npoetry install\npre-commit install\n```\n\n## Minimal working example\n\nThis section shows the steps to perform a resonator spectroscopy with Qibocal.\n### Write a runcard\nA runcard contains all the essential information to run a specific task.\nFor our purposes, we can use the following:\n```yml\nplatform: tii1q\n\nqubits: [0]\n\nformat: csv\n\nactions:\n   resonator_spectroscopy:\n     lowres_width: 5_000_000\n     lowres_step: 2_000_000\n     highres_width: 1_500_000\n     highres_step: 200_000\n     precision_width: 1_500_000\n     precision_step: 100_000\n     software_averages: 1\n     points: 5\n```\n### Run the routine\nTo run all the calibration routines specified in the ```runcard```, Qibocal uses the `qq` command\n```sh\nqq <runcard> -o <output_folder>\n```\nif ```<output_folder>``` is specified, the results will be saved in it, otherwise ```qq``` will automatically create a default folder containing the current date and the username.\n\n### Visualize the data\n\nQibocal gives the possibility to live-plotting with the `qq-live` command\n```sh\nqq-live <output_folder>\n```\n### Uploading reports to server\n\nIn order to upload the report to a centralized server, send to the server administrators your public ssh key (from the machine(s) you are planning to upload the report) and then use the `qq-upload <output_folder>` command. This program will upload your report to the server and generate an unique URL.\n\n## Contributing\n\nContributions, issues and feature requests are welcome!\nFeel free to check\n<a href="https://github.com/qiboteam/qibocal/issues"><img alt="GitHub issues" src="https://img.shields.io/github/issues-closed/qiboteam/qibocal"/></a>\n',
    'author': 'andrea-pasquale',
    'author_email': 'andreapasquale97@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/qiboteam/qibocal/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



"""Some tests for the Data and DataUnits class"""
import os
import shutil

import numpy as np
import pandas as pd
import pytest
from pint import DimensionalityError, UndefinedUnitError

from qibocal.data import AbstractData, Data, DataUnits


def random_data_units(length, options=None):
    data = DataUnits(options=options)
    for l in range(length):
        msr, i, q, phase = np.random.rand(4)
        pulse_sequence_result = {
            "MSR[V]": msr,
            "i[V]": i,
            "q[V]": q,
            "phase[rad]": phase,
        }
        add_options = {}
        if options is not None:
            for option in options:
                add_options[option] = l
        data.add({**pulse_sequence_result, **add_options})

    return data


def data_units_dummy(length, options=None):
    data = DataUnits(options=options)
    for l in range(length):
        pulse_sequence_result = {
            "MSR[V]": float(l),
            "i[V]": float(l),
            "q[V]": float(l),
            "phase[rad]": float(l),
        }
        add_options = {}
        if options is not None:
            for option in options:
                add_options[option] = float(l)
        data.add({**pulse_sequence_result, **add_options})

    return data


def random_data(length):
    data = Data()
    for i in range(length):
        data.add(
            {
                "int": int(i),
                "float": float(i),
                "string": str(f"hello{i}"),
                "bool": bool(i),
            }
        )
    return data


def data_dummy(length):
    data = Data()
    for i in range(length):
        data.add(
            {
                "num1": 0.0,
                "num2": 1.0,
            }
        )
    return data


def test_data_units_initialization():
    """Test DataUnits constructor"""
    data = DataUnits(name="data")
    assert data.name == "data"
    assert len(data.df.columns) == 4
    assert list(data.df.columns) == [  # pylint: disable=E1101
        "MSR",
        "i",
        "q",
        "phase",
    ]

    data1 = DataUnits(quantities={"attenuation": "dB"})
    assert len(data1.df.columns) == 5
    assert list(data1.df.columns) == [  # pylint: disable=E1101
        "attenuation",
        "MSR",
        "i",
        "q",
        "phase",
    ]

    data2 = DataUnits(quantities={"attenuation": "dB"}, options=["option1"])
    assert len(data2.df.columns) == 6
    assert list(data2.df.columns) == [  # pylint: disable=E1101
        "option1",
        "attenuation",
        "MSR",
        "i",
        "q",
        "phase",
    ]


def test_data_initialization():
    """Test initialization of class Data"""
    quantities_test = ["test"]
    data = Data(quantities=quantities_test)
    data.add(
        {
            "test": 0,
        }
    )
    assert data.quantities == quantities_test


def test_data_units_units():
    """Test units of measure in DataUnits"""
    data_units = DataUnits()
    assert data_units.df.MSR.values.units == "volt"

    data_units1 = DataUnits(quantities={"frequency": "Hz"})
    assert data_units1.df.frequency.values.units == "hertz"

    with pytest.raises(UndefinedUnitError):
        data_units2 = DataUnits(quantities={"fake_unit": "fake"})


def test_data_units_add():
    """Test add method of DataUnits"""
    data_units = random_data_units(5)
    assert len(data_units) == 5

    data_units1 = DataUnits(quantities={"attenuation": "dB"})
    msr, i, q, phase, att = np.random.rand(len(data_units1.df.columns))
    data_units1.add(
        {
            "MSR[V]": msr,
            "i[V]": i,
            "q[V]": q,
            "phase[rad]": phase,
            "attenuation[dB]": att,
        }
    )
    assert len(data_units1) == 1

    data_units1.add(
        {
            "MSR[V]": 0,
            "i[V]": 0.0,
            "q[V]": 0.0,
            "phase[rad]": 0,
            "attenuation[dB]": 1,
        }
    )
    assert len(data_units1) == 2

    data_units2 = DataUnits()
    msr, i, q, phase = np.random.rand(len(data_units2.df.columns))
    with pytest.raises(DimensionalityError):
        data_units2.add({"MSR[dB]": msr, "i[V]": i, "q[V]": q, "phase[rad]": phase})

    with pytest.raises(UndefinedUnitError):
        data_units2.add({"MSR[test]": msr, "i[V]": i, "q[V]": q, "phase[rad]": phase})

    data_units3 = random_data_units(10, options=["test"])
    assert len(data_units3) == 10


def test_data_add():
    """Test add method of Data class"""
    data = random_data(5)
    assert len(data) == 5
    data.add({"int": 123, "float": 123.456, "string": "123", "bool": True})
    assert len(data) == 6


def test_data__add__():
    data0 = Data()
    data0.add(
        {
            "col1": 0,
            "col2": 0,
        }
    )
    data1 = Data()
    data1.add(
        {
            "col1": 1,
            "col2": 1,
        }
    )
    data0.__add__(data1)
    data_results = Data()
    for i in [0, 1]:
        data_results.add(
            {
                "col1": int(i),
                "col2": int(i),
            }
        )
    assert data0.df.equals(data_results.df)


def test_data_units_load_data_from_dict():
    """Test set method of DataUnits class"""
    data_units = DataUnits()
    test = {
        "MSR[V]": [1, 2, 3],
        "i[V]": [3.0, 4.0, 5.0],
        "q[V]": np.array([3, 4, 5]),
        "phase[rad]": [6.0, 7.0, 8.0],
    }
    data_units.load_data_from_dict(test)
    assert len(data_units) == 3
    assert (data_units.get_values("MSR", "V") == [1, 2, 3]).all()
    assert (data_units.get_values("i", "V") == [3.0, 4.0, 5.0]).all()
    assert (data_units.get_values("q", "V") == [3, 4, 5]).all()
    assert (data_units.get_values("phase", "rad") == [6.0, 7.0, 8.0]).all()

    data_units1 = DataUnits(options=["option1", "option2"])
    test = {"option1": ["one", "two", "three"], "option2": [1, 2, 3]}
    data_units1.load_data_from_dict(test)
    assert len(data_units1) == 3
    assert (data_units1.get_values("option1") == ["one", "two", "three"]).all()
    assert (data_units1.get_values("option2") == [1, 2, 3]).all()


def test_df_data_units():
    """Test the method df in DataUnit class"""
    data = DataUnits()
    with pytest.raises(TypeError):
        data.df = 0


def test_data_load_data_from_dict():
    """Test set method of Data class"""
    data = random_data(5)
    test = {
        "int": [1, 2, 3],
        "float": [3.0, 4.0, 5.0],
        "string": ["one", "two", "three"],
        "bool": [True, False, True],
    }
    data.load_data_from_dict(test)
    assert len(data) == 3
    assert (data.get_values("int") == [1, 2, 3]).all()
    assert (data.get_values("float") == [3.0, 4.0, 5.0]).all()
    assert (data.get_values("string") == ["one", "two", "three"]).all()
    assert (data.get_values("bool") == [True, False, True]).all()


def test_get_values_data_units():
    """Test get_values method of DataUnits class"""
    data_units = random_data_units(5, options=["option"])

    assert (data_units.get_values("option") == data_units.df["option"]).all()
    assert (
        data_units.get_values("MSR", "uV")
        == data_units.df["MSR"].pint.to("uV").pint.magnitude
    ).all()


def test_get_values_data():
    """Test get_values method of Data class"""
    data = random_data(5)
    assert (data.get_values("int") == data.df["int"]).all()


def test_save_open_data_units_csv():
    """Test to_csv and load_data methods of DataUnits"""
    path = "test_folder/test_subfolder/test_routine"
    if not os.path.isdir(path):
        os.makedirs(path)
    data_units = data_units_dummy(5)
    data_units.to_csv(path)
    isExist = os.path.exists(f"{path}/{data_units.name}.csv")
    assert isExist is True
    with pytest.raises(ValueError):
        data_upload = DataUnits().load_data(
            "test_folder", "test_subfolder", "test_routine", "txt", "data"
        )
    data_upload = DataUnits().load_data(
        "test_folder", "test_subfolder", "test_routine", "csv", "data"
    )
    shutil.rmtree("test_folder")
    pd.testing.assert_frame_equal(data_upload.df, data_units.df)


def test_save_open_data_units_pickle():
    """Test to_csv and load_data methods of DataUnits"""
    folder = "test_folder/data/test_routine"
    if not os.path.isdir(folder):
        os.makedirs(folder)
    data_units = random_data_units(5)
    data_units.to_pickle(folder)
    data_upload = DataUnits().load_data(
        "test_folder", "data", "test_routine", "pickle", "data"
    )
    shutil.rmtree("test_folder")
    assert data_units.df.equals(data_upload.df)


def test_save_open_data_csv():
    """Test to_csv and load_data methods of Data"""
    path = "test_folder/data/test_routine"
    if not os.path.isdir(path):
        os.makedirs(path)
    data = data_dummy(5)
    data.to_csv(path)
    isExist = os.path.exists(f"{path}/{data.name}.csv")
    assert isExist is True
    with pytest.raises(ValueError):
        data_upload = Data().load_data(
            "test_folder", "data", "test_routine", "txt", "data"
        )
    data_upload = Data().load_data("test_folder", "data", "test_routine", "csv", "data")
    assert data.df.equals(data_upload.df)


def test_save_open_data_pickle():
    """Test to_pickle and load_data methods of Data"""
    folder = "test_folder/data/test_routine"
    if not os.path.isdir(folder):
        os.makedirs(folder)
    data = random_data(5)
    data.to_pickle(folder)
    data_upload = Data().load_data(
        "test_folder", "data", "test_routine", "pickle", "data"
    )
    shutil.rmtree("test_folder")
    assert data.df.equals(data_upload.df)


def test_load_data_from_dict_data_units():
    """Test load_data_from_dict method of DataUnits"""
    data_units = data_units_dummy(5)

    test_dict = {key: [0, 1, 2, 3] for key in data_units.df.columns}
    data_units.load_data_from_dict(test_dict)

    for column in data_units.df.columns:  # pylint: disable=E1101
        assert (data_units.get_values(column).to_numpy() == [0, 1, 2, 3]).all()


def test_load_data_from_dict_data():
    """Test load_data_from_dict method of Data"""
    data = data_dummy(5)
    test_dict = {key: [0, 1, 2, 3] for key in data.df.columns}
    data.load_data_from_dict(test_dict)

    for column in data.df.columns:  # pylint: disable=E1101
        assert (data.get_values(column) == [0, 1, 2, 3]).all()



import os
from colorsys import hls_to_rgb


def get_data_subfolders(folder):
    # iterate over multiple data folders
    subfolders = []
    for file in os.listdir(folder):
        d = os.path.join(folder, file)
        if os.path.isdir(d):
            subfolders.append(os.path.basename(d))

    return subfolders[::-1]


def get_color(number):
    return "rgb" + str(hls_to_rgb((0.75 - number * 3 / 20) % 1, 0.4, 0.75))


def get_color_state0(number):
    return "rgb" + str(hls_to_rgb((-0.35 - number * 9 / 20) % 1, 0.6, 0.75))


def get_color_state1(number):
    return "rgb" + str(hls_to_rgb((-0.02 - number * 9 / 20) % 1, 0.6, 0.75))



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import t1_fit


@plot("MSR vs Time", plots.t1_time_msr)
def t1(
    platform: AbstractPlatform,
    qubits: list,
    delay_before_readout_start,
    delay_before_readout_end,
    delay_before_readout_step,
    software_averages=1,
    points=10,
):
    r"""
    In a T1 experiment, we measure an excited qubit after a delay. Due to decoherence processes
    (e.g. amplitude damping channel), it is possible that, at the time of measurement, after the delay,
    the qubit will not be excited anymore. The larger the delay time is, the more likely is the qubit to
    fall to the ground state. The goal of the experiment is to characterize the decay rate of the qubit
    towards the ground state.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (list): List of target qubits to perform the action
        delay_before_readout_start (int): Initial time delay before ReadOut
        delay_before_readout_end (list): Maximum time delay before ReadOut
        delay_before_readout_step (int): Scan range step for the delay before ReadOut
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **wait[ns]**: Delay before ReadOut used in the current execution
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **labels[0]**: T1
            - **popt0**: p0
            - **popt1**: p1
            - **popt2**: p2
            - **qubit**: The qubit being tested
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment
    # RX - wait t - MZ
    qd_pulses = {}
    ro_pulses = {}
    sequence = PulseSequence()
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].duration
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # wait time before readout
    ro_wait_range = np.arange(
        delay_before_readout_start, delay_before_readout_end, delay_before_readout_step
    )

    # create a DataUnits object to store the MSR, phase, i, q and the delay time
    data = DataUnits(
        name=f"data", quantities={"wait": "ns"}, options=["qubit", "iteration"]
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        # sweep the parameter
        for wait in ro_wait_range:
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data
                # calculate and save fit
                yield t1_fit(
                    data,
                    x="wait[ns]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=["T1"],
                )

            for qubit in qubits:
                ro_pulses[qubit].start = qd_pulses[qubit].duration + wait

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            for ro_pulse in ro_pulses.values():
                # average msr, phase, i and q over the number of shots defined in the runcard
                r = results[ro_pulse.serial].to_dict()
                r.update(
                    {
                        "wait[ns]": wait,
                        "qubit": qubit,
                        "iteration": iteration,
                    }
                )
                data.add(r)
            count += 1
    yield data
    yield t1_fit(
        data,
        x="wait[ns]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=["T1"],
    )



import numpy as np
import pandas as pd
import pytest
from plotly.graph_objects import Figure
from qibo import gates
from qibo.noise import NoiseModel

from qibocal.calibrations.niGSC import XIdrb
from qibocal.calibrations.niGSC.basics import noisemodels, utils


@pytest.fixture
def depths():
    return [0, 1, 2, 3, 4, 5]


@pytest.fixture
def nshots():
    return 27


@pytest.mark.parametrize("nqubits", [1, 2])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("qubits", [[0], [1]])
def test_experiment(nqubits: int, depths: list, runs: int, nshots: int, qubits: list):
    if max(qubits) > nqubits - 1:
        qubits = [0, 1]
        with pytest.raises(ValueError):
            myfactory1 = XIdrb.ModuleFactory(2, list(depths) * runs, qubits)
    else:
        myfactory1 = XIdrb.ModuleFactory(nqubits, list(depths) * runs, qubits)
        myexperiment1 = XIdrb.ModuleExperiment(myfactory1, nshots=nshots)
        assert myexperiment1.name == "XIdRB"
        myexperiment1.perform(myexperiment1.execute)
        assert isinstance(myexperiment1.data, list)
        assert isinstance(myexperiment1.data[0], dict)
        for count, datarow in enumerate(myexperiment1.data):
            assert len(datarow.keys()) == 3
            assert isinstance(datarow["samples"], np.ndarray)
            assert len(datarow["samples"]) == nshots
            assert isinstance(datarow["depth"], int)
            assert datarow["depth"] == depths[count % len(depths)]
            assert np.array_equal(
                datarow["samples"],
                np.zeros(datarow["samples"].shape) + datarow["countX"] % 2,
            )
        assert isinstance(myexperiment1.dataframe, pd.DataFrame)


@pytest.mark.parametrize("nqubits", [1, 3])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("noise_params", [[0.1, 0.1, 0.1], [0.02, 0.3, 0.07]])
@pytest.mark.parametrize("qubits", [[0], [1]])
def test_experiment_withnoise(
    nqubits: int, depths: list, runs: int, qubits: list, noise_params: list
):
    nshots = 100
    if max(qubits) > nqubits - 1:
        pass
    else:
        # Build the noise model.
        noise = noisemodels.PauliErrorOnX(*noise_params)
        # Test exectue an experiment.
        myfactory1 = XIdrb.ModuleFactory(nqubits, list(depths) * runs, qubits)
        myfaultyexperiment = XIdrb.ModuleExperiment(
            myfactory1, nshots=nshots, noise_model=noise
        )
        myfaultyexperiment.perform(myfaultyexperiment.execute)
        assert isinstance(myfaultyexperiment.data, list)
        assert isinstance(myfaultyexperiment.data[0], dict)
        for count, datarow in enumerate(myfaultyexperiment.data):
            assert len(datarow.keys()) == 3
            assert isinstance(datarow["samples"], np.ndarray)
            assert len(datarow["samples"]) == nshots
            assert isinstance(datarow["depth"], int)
            assert datarow["depth"] == depths[count % len(depths)]
            if not datarow["countX"]:
                assert np.array_equal(
                    datarow["samples"], np.zeros(datarow["samples"].shape)
                )
            else:
                theor_outcome = datarow["countX"] % 2
                assert not np.array_equal(
                    datarow["samples"],
                    np.zeros(datarow["samples"].shape) + theor_outcome,
                )
        assert isinstance(myfaultyexperiment.dataframe, pd.DataFrame)


@pytest.mark.parametrize("nqubits", [1, 4])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("qubits", [[0], [2]])
def test_post_processing(
    nqubits: int, depths: list, runs: int, nshots: int, qubits: list
):
    if max(qubits) > nqubits - 1:
        pass
    else:
        noise_params = [0.01, 0.3, 0.14]
        # Build the noise model.
        noise = noisemodels.PauliErrorOnX(*noise_params)
        # Test exectue an experiment.
        myfactory1 = XIdrb.ModuleFactory(nqubits, list(depths) * runs, qubits)
        myfaultyexperiment = XIdrb.ModuleExperiment(
            myfactory1, nshots=nshots, noise_model=noise
        )
        myfaultyexperiment.perform(myfaultyexperiment.execute)
        XIdrb.post_processing_sequential(myfaultyexperiment)
        aggr_df = XIdrb.get_aggregational_data(myfaultyexperiment)
        assert len(aggr_df) == 1 and aggr_df.index[0] == "filter"
        assert "depth" in aggr_df.columns
        assert "data" in aggr_df.columns
        assert "2sigma" in aggr_df.columns
        assert "fit_func" in aggr_df.columns
        assert "popt_real" in aggr_df.columns
        assert "perr" in aggr_df.columns


@pytest.mark.parametrize("nqubits", [1, 5])
@pytest.mark.parametrize("runs", [1, 3])
@pytest.mark.parametrize("qubits", [[0], [2]])
def test_build_report(depths: list, nshots: int, nqubits: int, runs: int, qubits: list):
    if max(qubits) > nqubits - 1:
        pass
    else:
        noise_params = [0.01, 0.1, 0.05]
        # Build the noise model.
        noise = noisemodels.PauliErrorOnX(*noise_params)
        # Test exectue an experiment.
        myfactory1 = XIdrb.ModuleFactory(nqubits, depths * runs, qubits)
        myfaultyexperiment = XIdrb.ModuleExperiment(
            myfactory1, nshots=nshots, noise_model=noise
        )
        myfaultyexperiment.perform(myfaultyexperiment.execute)
        XIdrb.post_processing_sequential(myfaultyexperiment)
        aggr_df = XIdrb.get_aggregational_data(myfaultyexperiment)
        report_figure = XIdrb.build_report(myfaultyexperiment, aggr_df)
        assert isinstance(report_figure, Figure)



import numpy as np
import pytest
import qibo

from qibocal.calibrations.niGSC.basics import noisemodels


def test_PauliErrorOnUnitary():
    def test_model(pauli_onU_error):
        assert isinstance(pauli_onU_error, qibo.noise.NoiseModel)
        errorkeys = pauli_onU_error.errors.keys()
        assert len(errorkeys) == 1 and list(errorkeys)[0] == qibo.gates.gates.Unitary
        error = pauli_onU_error.errors[qibo.gates.gates.Unitary][0]
        assert isinstance(error, qibo.noise.PauliError)
        assert len(error.options) == 3 and np.sum(error.options) < 1

    noise_model1 = noisemodels.PauliErrorOnUnitary()
    test_model(noise_model1)
    noise_model2 = noisemodels.PauliErrorOnUnitary(0.1, 0.1, 0.1)
    test_model(noise_model2)
    noise_model3 = noisemodels.PauliErrorOnUnitary(None)
    test_model(noise_model3)
    with pytest.raises(ValueError):
        noise_model4 = noisemodels.PauliErrorOnUnitary(0.1, 0.2)


def test_PauliErrorOnX():
    def test_model(pauli_onX_error):
        assert isinstance(pauli_onX_error, qibo.noise.NoiseModel)
        errorkeys = pauli_onX_error.errors.keys()
        assert len(errorkeys) == 1 and list(errorkeys)[0] == qibo.gates.gates.X
        error = pauli_onX_error.errors[qibo.gates.gates.X][0]
        assert isinstance(error, qibo.noise.PauliError)
        assert len(error.options) == 3 and np.sum(error.options) < 1

    noise_model1 = noisemodels.PauliErrorOnX()
    test_model(noise_model1)
    noise_model2 = noisemodels.PauliErrorOnX(0.1, 0.1, 0.1)
    test_model(noise_model2)
    noise_model3 = noisemodels.PauliErrorOnX(None)
    test_model(noise_model3)
    with pytest.raises(ValueError):
        noise_model4 = noisemodels.PauliErrorOnX(0.1, 0.2)



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import calibrate_qubit_states_fit


@plot("Qubit States", plots.qubit_states)
def calibrate_qubit_states(
    platform: AbstractPlatform,
    qubits: dict,
    nshots,
    points=10,
):
    """
    Method which implements the state's calibration of a chosen qubit. Two analogous tests are performed
    for calibrate the ground state and the excited state of the oscillator.
    The subscripts `exc` and `gnd` will represent the excited state |1> and the ground state |0>.

    Args:
        platform (:class:`qibolab.platforms.abstract.AbstractPlatform`): custom abstract platform on which we perform the calibration.
        qubits (dict): Dict of target Qubit objects to perform the action
        nshots (int): number of times the pulse sequence will be repeated.
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **iteration[dimensionless]**: Execution number
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create two sequences of pulses:
    # state0_sequence: I  - MZ
    # state1_sequence: RX - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    state0_sequence = PulseSequence()
    state1_sequence = PulseSequence()

    RX_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        RX_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX_pulses[qubit].duration
        )

        state0_sequence.add(ro_pulses[qubit])
        state1_sequence.add(RX_pulses[qubit])
        state1_sequence.add(ro_pulses[qubit])

    # create a DataUnits object to store the results
    data = DataUnits(name="data", options=["qubit", "iteration", "state"])

    # execute the first pulse sequence
    state0_results = platform.execute_pulse_sequence(state0_sequence, nshots=nshots)

    # retrieve and store the results for every qubit
    for ro_pulse in ro_pulses.values():
        r = state0_results[ro_pulse.serial].to_dict(average=False)
        r.update(
            {
                "qubit": [ro_pulse.qubit] * nshots,
                "iteration": np.arange(nshots),
                "state": [0] * nshots,
            }
        )
        data.add_data_from_dict(r)

    # execute the second pulse sequence
    state1_results = platform.execute_pulse_sequence(state1_sequence, nshots=nshots)

    # retrieve and store the results for every qubit
    for ro_pulse in ro_pulses.values():
        r = state1_results[ro_pulse.serial].to_dict(average=False)
        r.update(
            {
                "qubit": [ro_pulse.qubit] * nshots,
                "iteration": np.arange(nshots),
                "state": [1] * nshots,
            }
        )
        data.add_data_from_dict(r)
    # finally, save the remaining data and the fits
    yield data
    yield calibrate_qubit_states_fit(
        data, x="i[V]", y="q[V]", nshots=nshots, qubits=qubits
    )
